"""
This file contains all the utilities used in that project. They are classified in 5 categories:
1- loading/saving functions:
    -load_volume
    -save_volume
    -get_volume_info
    -get_list_labels
    -load_array_if_path
    -write_pickle
    -read_pickle
    -write_model_summary
2- reformatting functions
    -reformat_to_list
    -reformat_to_n_channels_array
3- path related functions
    -list_images_in_folder
    -list_files
    -list_subfolders
    -strip_extension
    -strip_suffix
    -mkdir
    -mkcmd
4- shape-related functions
    -get_dims
    -get_resample_shape
    -add_axis
    -get_padding_margin
5- build affine matrices/tensors
    -create_affine_transformation_matrix
    -sample_affine_transform
    -create_rotation_transform
    -create_shearing_transform
6- miscellaneous
    -infer
    -LoopInfo
    -get_mapping_lut
    -build_training_generator
    -find_closest_number_divisible_by_m
    -build_binary_structure
    -draw_value_from_distribution
    -build_exp


If you use this code, please cite the first SynthSeg paper:
https://github.com/BBillot/lab2im/blob/master/bibtex.bib

Copyright 2020 Benjamin Billot

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software distributed under the License is
distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
implied. See the License for the specific language governing permissions and limitations under the
License.
"""


import os
import glob
import math
import time
import pickle
import numpy as np
import nibabel as nib
import tensorflow as tf
import keras.layers as KL
import keras.backend as K
from datetime import timedelta
from scipy.ndimage.morphology import distance_transform_edt


# ---------------------------------------------- loading/saving functions ----------------------------------------------


def load_volume(path_volume, im_only=True, squeeze=True, dtype=None, aff_ref=None):
    """
    Load volume file.
    :param path_volume: path of the volume to load. Can either be a nii, nii.gz, mgz, or npz format.
    If npz format, 1) the variable name is assumed to be 'vol_data',
    2) the volume is associated with an identity affine matrix and blank header.
    :param im_only: (optional) if False, the function also returns the affine matrix and header of the volume.
    :param squeeze: (optional) whether to squeeze the volume when loading.
    :param dtype: (optional) if not None, convert the loaded volume to this numpy dtype.
    :param aff_ref: (optional) If not None, the loaded volume is aligned to this affine matrix.
    The returned affine matrix is also given in this new space. Must be a numpy array of dimension 4x4.
    :return: the volume, with corresponding affine matrix and header if im_only is False.
    """
    assert path_volume.endswith(('.nii', '.nii.gz', '.mgz', '.npz')), 'Unknown data file: %s' % path_volume

    if path_volume.endswith(('.nii', '.nii.gz', '.mgz')):
        x = nib.load(path_volume)
        if squeeze:
            volume = np.squeeze(x.get_fdata())
        else:
            volume = x.get_fdata()
        aff = x.affine
        header = x.header
    else:  # npz
        volume = np.load(path_volume)['vol_data']
        if squeeze:
            volume = np.squeeze(volume)
        aff = np.eye(4)
        header = nib.Nifti1Header()
    if dtype is not None:
        if 'int' in dtype:
            volume = np.round(volume)
        volume = volume.astype(dtype=dtype)

    # align image to reference affine matrix
    if aff_ref is not None:
        from ext.lab2im import edit_volumes  # the import is done here to avoid import loops
        n_dims, _ = get_dims(list(volume.shape), max_channels=10)
        volume, aff = edit_volumes.align_volume_to_ref(volume, aff, aff_ref=aff_ref, return_aff=True, n_dims=n_dims)

    if im_only:
        return volume
    else:
        return volume, aff, header


def save_volume(volume, aff, header, path, res=None, dtype=None, n_dims=3):
    """
    Save a volume.
    :param volume: volume to save
    :param aff: affine matrix of the volume to save. If aff is None, the volume is saved with an identity affine matrix.
    aff can also be set to 'FS', in which case the volume is saved with the affine matrix of FreeSurfer outputs.
    :param header: header of the volume to save. If None, the volume is saved with a blank header.
    :param path: path where to save the volume.
    :param res: (optional) update the resolution in the header before saving the volume.
    :param dtype: (optional) numpy dtype for the saved volume.
    :param n_dims: (optional) number of dimensions, to avoid confusion in multi-channel case. Default is None, where
    n_dims is automatically inferred.
    """

    mkdir(os.path.dirname(path))
    if '.npz' in path:
        np.savez_compressed(path, vol_data=volume)
    else:
        if header is None:
            header = nib.Nifti1Header()
        if isinstance(aff, str):
            if aff == 'FS':
                aff = np.array([[-1, 0, 0, 0], [0, 0, 1, 0], [0, -1, 0, 0], [0, 0, 0, 1]])
        elif aff is None:
            aff = np.eye(4)
        if dtype is not None:
            if 'int' in dtype:
                volume = np.round(volume)
            volume = volume.astype(dtype=dtype)
            nifty = nib.Nifti1Image(volume, aff, header)
            nifty.set_data_dtype(dtype)
        else:
            nifty = nib.Nifti1Image(volume, aff, header)
        if res is not None:
            if n_dims is None:
                n_dims, _ = get_dims(volume.shape)
            res = reformat_to_list(res, length=n_dims, dtype=None)
            nifty.header.set_zooms(res)
        nib.save(nifty, path)


def get_volume_info(path_volume, return_volume=False, aff_ref=None, max_channels=10):
    """
    Gather information about a volume: shape, affine matrix, number of dimensions and channels, header, and resolution.
    :param path_volume: path of the volume to get information form.
    :param return_volume: (optional) whether to return the volume along with the information.
    :param aff_ref: (optional) If not None, the loaded volume is aligned to this affine matrix.
    All info relative to the volume is then given in this new space. Must be a numpy array of dimension 4x4.
    :param max_channels: maximum possible number of channels for the input volume.
    :return: volume (if return_volume is true), and corresponding info. If aff_ref is not None, the returned aff is
    the original one, i.e. the affine of the image before being aligned to aff_ref.
    """
    # read image
    im, aff, header = load_volume(path_volume, im_only=False)

    # understand if image is multichannel
    im_shape = list(im.shape)
    n_dims, n_channels = get_dims(im_shape, max_channels=max_channels)
    im_shape = im_shape[:n_dims]

    # get labels res
    if '.nii' in path_volume:
        data_res = np.array(header['pixdim'][1:n_dims + 1])
    elif '.mgz' in path_volume:
        data_res = np.array(header['delta'])  # mgz image
    else:
        data_res = np.array([1.0] * n_dims)

    # align to given affine matrix
    if aff_ref is not None:
        from ext.lab2im import edit_volumes  # the import is done here to avoid import loops
        ras_axes = edit_volumes.get_ras_axes(aff, n_dims=n_dims)
        ras_axes_ref = edit_volumes.get_ras_axes(aff_ref, n_dims=n_dims)
        im = edit_volumes.align_volume_to_ref(im, aff, aff_ref=aff_ref, n_dims=n_dims)
        im_shape = np.array(im_shape)
        data_res = np.array(data_res)
        im_shape[ras_axes_ref] = im_shape[ras_axes]
        data_res[ras_axes_ref] = data_res[ras_axes]
        im_shape = im_shape.tolist()

    # return info
    if return_volume:
        return im, im_shape, aff, n_dims, n_channels, header, data_res
    else:
        return im_shape, aff, n_dims, n_channels, header, data_res


def get_list_labels(label_list=None, labels_dir=None, save_label_list=None, FS_sort=False):
    """This function reads or computes a list of all label values used in a set of label maps.
    It can also sort all labels according to FreeSurfer lut.
    :param label_list: (optional) already computed label_list. Can be a sequence, a 1d numpy array, or the path to
    a numpy 1d array.
    :param labels_dir: (optional) if path_label_list is None, the label list is computed by reading all the label maps
    in the given folder. Can also be the path to a single label map.
    :param save_label_list: (optional) path where to save the label list.
    :param FS_sort: (optional) whether to sort label values according to the FreeSurfer classification.
    If true, the label values will be ordered as follows: neutral labels first (i.e. non-sided), left-side labels,
    and right-side labels. If FS_sort is True, this function also returns the number of neutral labels in label_list.
    :return: the label list (numpy 1d array), and the number of neutral (i.e. non-sided) labels if FS_sort is True.
    If one side of the brain is not represented at all in label_list, all labels are considered as neutral, and
    n_neutral_labels = len(label_list).
    """

    # load label list if previously computed
    if label_list is not None:
        label_list = np.array(reformat_to_list(label_list, load_as_numpy=True, dtype='int'))

    # compute label list from all label files
    elif labels_dir is not None:
        print('Compiling list of unique labels')
        # go through all labels files and compute unique list of labels
        labels_paths = list_images_in_folder(labels_dir)
        label_list = np.empty(0)
        loop_info = LoopInfo(len(labels_paths), 10, 'processing', print_time=True)
        for lab_idx, path in enumerate(labels_paths):
            loop_info.update(lab_idx)
            y = load_volume(path, dtype='int32')
            y_unique = np.unique(y)
            label_list = np.unique(np.concatenate((label_list, y_unique))).astype('int')

    else:
        raise Exception('either label_list, path_label_list or labels_dir should be provided')

    # sort labels in neutral/left/right according to FS labels
    n_neutral_labels = 0
    if FS_sort:
        neutral_FS_labels = [0, 14, 15, 16, 21, 22, 23, 24, 72, 77, 80, 85, 100, 101, 102, 103, 104, 105, 106, 107, 108,
                             109, 165, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210,
                             251, 252, 253, 254, 255, 258, 259, 260, 331, 332, 333, 334, 335, 336, 337, 338, 339, 340,
                             502, 506, 507, 508, 509, 511, 512, 514, 515, 516, 517, 530,
                             531, 532, 533, 534, 535, 536, 537]
        neutral = list()
        left = list()
        right = list()
        for la in label_list:
            if la in neutral_FS_labels:
                if la not in neutral:
                    neutral.append(la)
            elif (0 < la < 14) | (16 < la < 21) | (24 < la < 40) | (135 < la < 139) | (1000 <= la <= 1035) | \
                    (la == 865) | (20100 < la < 20110):
                if la not in left:
                    left.append(la)
            elif (39 < la < 72) | (162 < la < 165) | (2000 <= la <= 2035) | (20000 < la < 20010) | (la == 139) | \
                    (la == 866):
                if la not in right:
                    right.append(la)
            else:
                raise Exception('label {} not in our current FS classification, '
                                'please update get_list_labels in utils.py'.format(la))
        label_list = np.concatenate([sorted(neutral), sorted(left), sorted(right)])
        if ((len(left) > 0) & (len(right) > 0)) | ((len(left) == 0) & (len(right) == 0)):
            n_neutral_labels = len(neutral)
        else:
            n_neutral_labels = len(label_list)

    # save labels if specified
    if save_label_list is not None:
        np.save(save_label_list, np.int32(label_list))

    if FS_sort:
        return np.int32(label_list), n_neutral_labels
    else:
        return np.int32(label_list), None


def load_array_if_path(var, load_as_numpy=True):
    """If var is a string and load_as_numpy is True, this function loads the array writen at the path indicated by var.
    Otherwise it simply returns var as it is."""
    if (isinstance(var, str)) & load_as_numpy:
        assert os.path.isfile(var), 'No such path: %s' % var
        var = np.load(var)
    return var


def write_pickle(filepath, obj):
    """ write a python object with a pickle at a given path"""
    with open(filepath, 'wb') as file:
        pickler = pickle.Pickler(file)
        pickler.dump(obj)


def read_pickle(filepath):
    """ read a python object with a pickle"""
    with open(filepath, 'rb') as file:
        unpickler = pickle.Unpickler(file)
        return unpickler.load()


def write_model_summary(model, filepath='./model_summary.txt', line_length=150):
    """Write the summary of a keras model at a given path, with a given length for each line"""
    with open(filepath, 'w') as fh:
        model.summary(print_fn=lambda x: fh.write(x + '\n'), line_length=line_length)


# ----------------------------------------------- reformatting functions -----------------------------------------------


def reformat_to_list(var, length=None, load_as_numpy=False, dtype=None):
    """This function takes a variable and reformat it into a list of desired
    length and type (int, float, bool, str).
    If variable is a string, and load_as_numpy is True, it will be loaded as a numpy array.
    If variable is None, this function returns None.
    :param var: a str, int, float, list, tuple, or numpy array
    :param length: (optional) if var is a single item, it will be replicated to a list of this length
    :param load_as_numpy: (optional) whether var is the path to a numpy array
    :param dtype: (optional) convert all item to this type. Can be 'int', 'float', 'bool', or 'str'
    :return: reformatted list
    """

    # convert to list
    if var is None:
        return None
    var = load_array_if_path(var, load_as_numpy=load_as_numpy)
    if isinstance(var, (int, float, np.int32, np.int64, np.float32, np.float64)):
        var = [var]
    elif isinstance(var, tuple):
        var = list(var)
    elif isinstance(var, np.ndarray):
        if var.shape == (1,):
            var = [var[0]]
        else:
            var = np.squeeze(var).tolist()
    elif isinstance(var, str):
        var = [var]
    elif isinstance(var, bool):
        var = [var]
    if isinstance(var, list):
        if length is not None:
            if len(var) == 1:
                var = var * length
            elif len(var) != length:
                raise ValueError('if var is a list/tuple/numpy array, it should be of length 1 or {0}, '
                                 'had {1}'.format(length, var))
    else:
        raise TypeError('var should be an int, float, tuple, list, numpy array, or path to numpy array')

    # convert items type
    if dtype is not None:
        if dtype == 'int':
            var = [int(v) for v in var]
        elif dtype == 'float':
            var = [float(v) for v in var]
        elif dtype == 'bool':
            var = [bool(v) for v in var]
        elif dtype == 'str':
            var = [str(v) for v in var]
        else:
            raise ValueError("dtype should be 'str', 'float', 'int', or 'bool'; had {}".format(dtype))
    return var


def reformat_to_n_channels_array(var, n_dims=3, n_channels=1):
    """This function takes an int, float, list or tuple and reformat it to an array of shape (n_channels, n_dims).
    If resolution is a str, it will be assumed to be the path of a numpy array.
    If resolution is a numpy array, it will be checked to have shape (n_channels, n_dims).
    Finally if resolution is None, this function returns None as well."""
    if var is None:
        return [None] * n_channels
    if isinstance(var, str):
        var = np.load(var)
    # convert to numpy array
    if isinstance(var, (int, float, list, tuple)):
        var = reformat_to_list(var, n_dims)
        var = np.tile(np.array(var), (n_channels, 1))
    # check shape if numpy array
    elif isinstance(var, np.ndarray):
        if n_channels == 1:
            var = var.reshape((1, n_dims))
        else:
            if np.squeeze(var).shape == (n_dims,):
                var = np.tile(var.reshape((1, n_dims)), (n_channels, 1))
            elif var.shape != (n_channels, n_dims):
                raise ValueError('if array, var should be {0} or {1}'.format((1, n_dims), (n_channels, n_dims)))
    else:
        raise TypeError('var should be int, float, list, tuple or ndarray')
    return np.round(var, 3)


# ----------------------------------------------- path-related functions -----------------------------------------------


def list_images_in_folder(path_dir, include_single_image=True, check_if_empty=True):
    """List all files with extension nii, nii.gz, mgz, or npz within a folder."""
    basename = os.path.basename(path_dir)
    if include_single_image & \
            (('.nii.gz' in basename) | ('.nii' in basename) | ('.mgz' in basename) | ('.npz' in basename)):
        assert os.path.isfile(path_dir), 'file %s does not exist' % path_dir
        list_images = [path_dir]
    else:
        if os.path.isdir(path_dir):
            list_images = sorted(glob.glob(os.path.join(path_dir, '*nii.gz')) +
                                 glob.glob(os.path.join(path_dir, '*nii')) +
                                 glob.glob(os.path.join(path_dir, '*.mgz')) +
                                 glob.glob(os.path.join(path_dir, '*.npz')))
        else:
            raise Exception('Folder does not exist: %s' % path_dir)
        if check_if_empty:
            assert len(list_images) > 0, 'no .nii, .nii.gz, .mgz or .npz image could be found in %s' % path_dir
    return list_images


def list_files(path_dir, whole_path=True, expr=None, cond_type='or'):
    """This function returns a list of files contained in a folder, with possible regexp.
    :param path_dir: path of a folder
    :param whole_path: (optional) whether to return whole path or just the filenames.
    :param expr: (optional) regexp for files to list. Can be a str or a list of str.
    :param cond_type: (optional) if exp is a list, specify the logical link between expressions in exp.
    Can be 'or', or 'and'.
    :return: a list of files
    """
    assert isinstance(whole_path, bool), "whole_path should be bool"
    assert cond_type in ['or', 'and'], "cond_type should be either 'or', or 'and'"
    if whole_path:
        files_list = sorted([os.path.join(path_dir, f) for f in os.listdir(path_dir)
                             if os.path.isfile(os.path.join(path_dir, f))])
    else:
        files_list = sorted([f for f in os.listdir(path_dir) if os.path.isfile(os.path.join(path_dir, f))])
    if expr is not None:  # assumed to be either str or list of str
        if isinstance(expr, str):
            expr = [expr]
        elif not isinstance(expr, (list, tuple)):
            raise Exception("if specified, 'expr' should be a string or list of strings.")
        matched_list_files = list()
        for match in expr:
            tmp_matched_files_list = sorted([f for f in files_list if match in os.path.basename(f)])
            if cond_type == 'or':
                files_list = [f for f in files_list if f not in tmp_matched_files_list]
                matched_list_files += tmp_matched_files_list
            elif cond_type == 'and':
                files_list = tmp_matched_files_list
                matched_list_files = tmp_matched_files_list
        files_list = sorted(matched_list_files)
    return files_list


def list_subfolders(path_dir, whole_path=True, expr=None, cond_type='or'):
    """This function returns a list of subfolders contained in a folder, with possible regexp.
    :param path_dir: path of a folder
    :param whole_path: (optional) whether to return whole path or just the subfolder names.
    :param expr: (optional) regexp for files to list. Can be a str or a list of str.
    :param cond_type: (optional) if exp is a list, specify the logical link between expressions in exp.
    Can be 'or', or 'and'.
    :return: a list of subfolders
    """
    assert isinstance(whole_path, bool), "whole_path should be bool"
    assert cond_type in ['or', 'and'], "cond_type should be either 'or', or 'and'"
    if whole_path:
        subdirs_list = sorted([os.path.join(path_dir, f) for f in os.listdir(path_dir)
                               if os.path.isdir(os.path.join(path_dir, f))])
    else:
        subdirs_list = sorted([f for f in os.listdir(path_dir) if os.path.isdir(os.path.join(path_dir, f))])
    if expr is not None:  # assumed to be either str or list of str
        if isinstance(expr, str):
            expr = [expr]
        elif not isinstance(expr, (list, tuple)):
            raise Exception("if specified, 'expr' should be a string or list of strings.")
        matched_list_subdirs = list()
        for match in expr:
            tmp_matched_list_subdirs = sorted([f for f in subdirs_list if match in os.path.basename(f)])
            if cond_type == 'or':
                subdirs_list = [f for f in subdirs_list if f not in tmp_matched_list_subdirs]
                matched_list_subdirs += tmp_matched_list_subdirs
            elif cond_type == 'and':
                subdirs_list = tmp_matched_list_subdirs
                matched_list_subdirs = tmp_matched_list_subdirs
        subdirs_list = sorted(matched_list_subdirs)
    return subdirs_list


def get_image_extension(path):
    name = os.path.basename(path)
    if name[-7:] == '.nii.gz':
        return 'nii.gz'
    elif name[-4:] == '.mgz':
        return 'mgz'
    elif name[-4:] == '.nii':
        return 'nii'
    elif name[-4:] == '.npz':
        return 'npz'


def strip_extension(path):
    """Strip classical image extensions (.nii.gz, .nii, .mgz, .npz) from a filename."""
    return path.replace('.nii.gz', '').replace('.nii', '').replace('.mgz', '').replace('.npz', '')


def strip_suffix(path):
    """Strip classical image suffix from a filename."""
    path = path.replace('_aseg', '')
    path = path.replace('aseg', '')
    path = path.replace('.aseg', '')
    path = path.replace('_aseg_1', '')
    path = path.replace('_aseg_2', '')
    path = path.replace('aseg_1_', '')
    path = path.replace('aseg_2_', '')
    path = path.replace('_orig', '')
    path = path.replace('orig', '')
    path = path.replace('.orig', '')
    path = path.replace('_norm', '')
    path = path.replace('norm', '')
    path = path.replace('.norm', '')
    path = path.replace('_talairach', '')
    path = path.replace('GSP_FS_4p5', 'GSP')
    path = path.replace('.nii_crispSegmentation', '')
    path = path.replace('_crispSegmentation', '')
    path = path.replace('_seg', '')
    path = path.replace('.seg', '')
    path = path.replace('seg', '')
    path = path.replace('_seg_1', '')
    path = path.replace('_seg_2', '')
    path = path.replace('seg_1_', '')
    path = path.replace('seg_2_', '')
    return path


def mkdir(path_dir):
    """Recursively creates the current dir as well as its parent folders if they do not already exist."""
    if path_dir[-1] == '/':
        path_dir = path_dir[:-1]
    if not os.path.isdir(path_dir):
        list_dir_to_create = [path_dir]
        while not os.path.isdir(os.path.dirname(list_dir_to_create[-1])):
            list_dir_to_create.append(os.path.dirname(list_dir_to_create[-1]))
        for dir_to_create in reversed(list_dir_to_create):
            os.mkdir(dir_to_create)


def mkcmd(*args):
    """Creates terminal command with provided inputs.
    Example: mkcmd('mv', 'source', 'dest') will give 'mv source dest'."""
    return ' '.join([str(arg) for arg in args])


# ---------------------------------------------- shape-related functions -----------------------------------------------


def get_dims(shape, max_channels=10):
    """Get the number of dimensions and channels from the shape of an array.
    The number of dimensions is assumed to be the length of the shape, as long as the shape of the last dimension is
    inferior or equal to max_channels (default 3).
    :param shape: shape of an array. Can be a sequence or a 1d numpy array.
    :param max_channels: maximum possible number of channels.
    :return: the number of dimensions and channels associated with the provided shape.
    example 1: get_dims([150, 150, 150], max_channels=10) = (3, 1)
    example 2: get_dims([150, 150, 150, 3], max_channels=10) = (3, 3)
    example 3: get_dims([150, 150, 150, 15], max_channels=10) = (4, 1), because 5>3"""
    if shape[-1] <= max_channels:
        n_dims = len(shape) - 1
        n_channels = shape[-1]
    else:
        n_dims = len(shape)
        n_channels = 1
    return n_dims, n_channels


def get_resample_shape(patch_shape, factor, n_channels=None):
    """Compute the shape of a resampled array given a shape factor.
    :param patch_shape: size of the initial array (without number of channels).
    :param factor: resampling factor. Can be a number, sequence, or 1d numpy array.
    :param n_channels: (optional) if not None, add a number of channel at the end of the computed shape.
    :return: list containing the shape of the input array after being resampled by the given factor.
    """
    factor = reformat_to_list(factor, length=len(patch_shape))
    shape = [math.ceil(patch_shape[i] * factor[i]) for i in range(len(patch_shape))]
    if n_channels is not None:
        shape += [n_channels]
    return shape


def add_axis(x, axis=0):
    """Add axis to a numpy array.
    :param x: input array
    :param axis: index of the new axis to add. Can also be a list of indices to add several axes at the same time."""
    axis = reformat_to_list(axis)
    for ax in axis:
        x = np.expand_dims(x, axis=ax)
    return x


def get_padding_margin(cropping, loss_cropping):
    """Compute padding margin"""
    if (cropping is not None) & (loss_cropping is not None):
        cropping = reformat_to_list(cropping)
        loss_cropping = reformat_to_list(loss_cropping)
        n_dims = max(len(cropping), len(loss_cropping))
        cropping = reformat_to_list(cropping, length=n_dims)
        loss_cropping = reformat_to_list(loss_cropping, length=n_dims)
        padding_margin = [int((cropping[i] - loss_cropping[i]) / 2) for i in range(n_dims)]
        if len(padding_margin) == 1:
            padding_margin = padding_margin[0]
    else:
        padding_margin = None
    return padding_margin


# -------------------------------------------- build affine matrices/tensors -------------------------------------------


def create_affine_transformation_matrix(n_dims, scaling=None, rotation=None, shearing=None, translation=None):
    """Create a 4x4 affine transformation matrix from specified values
    :param n_dims: integer, can either be 2 or 3.
    :param scaling: list of 3 scaling values
    :param rotation: list of 3 angles (degrees) for rotations around 1st, 2nd, 3rd axis
    :param shearing: list of 6 shearing values
    :param translation: list of 3 values
    :return: 4x4 numpy matrix
    """

    T_scaling = np.eye(n_dims + 1)
    T_shearing = np.eye(n_dims + 1)
    T_translation = np.eye(n_dims + 1)

    if scaling is not None:
        T_scaling[np.arange(n_dims + 1), np.arange(n_dims + 1)] = np.append(scaling, 1)

    if shearing is not None:
        shearing_index = np.ones((n_dims + 1, n_dims + 1), dtype='bool')
        shearing_index[np.eye(n_dims + 1, dtype='bool')] = False
        shearing_index[-1, :] = np.zeros((n_dims + 1))
        shearing_index[:, -1] = np.zeros((n_dims + 1))
        T_shearing[shearing_index] = shearing

    if translation is not None:
        T_translation[np.arange(n_dims), n_dims * np.ones(n_dims, dtype='int')] = translation

    if n_dims == 2:
        if rotation is None:
            rotation = np.zeros(1)
        else:
            rotation = np.asarray(rotation) * (math.pi / 180)
        T_rot = np.eye(n_dims + 1)
        T_rot[np.array([0, 1, 0, 1]), np.array([0, 0, 1, 1])] = [np.cos(rotation[0]), np.sin(rotation[0]),
                                                                 np.sin(rotation[0]) * -1, np.cos(rotation[0])]
        return T_translation @ T_rot @ T_shearing @ T_scaling

    else:

        if rotation is None:
            rotation = np.zeros(n_dims)
        else:
            rotation = np.asarray(rotation) * (math.pi / 180)
        T_rot1 = np.eye(n_dims + 1)
        T_rot1[np.array([1, 2, 1, 2]), np.array([1, 1, 2, 2])] = [np.cos(rotation[0]), np.sin(rotation[0]),
                                                                  np.sin(rotation[0]) * -1, np.cos(rotation[0])]
        T_rot2 = np.eye(n_dims + 1)
        T_rot2[np.array([0, 2, 0, 2]), np.array([0, 0, 2, 2])] = [np.cos(rotation[1]), np.sin(rotation[1]) * -1,
                                                                  np.sin(rotation[1]), np.cos(rotation[1])]
        T_rot3 = np.eye(n_dims + 1)
        T_rot3[np.array([0, 1, 0, 1]), np.array([0, 0, 1, 1])] = [np.cos(rotation[2]), np.sin(rotation[2]),
                                                                  np.sin(rotation[2]) * -1, np.cos(rotation[2])]
        return T_translation @ T_rot3 @ T_rot2 @ T_rot1 @ T_shearing @ T_scaling


def sample_affine_transform(batchsize,
                            n_dims,
                            rotation_bounds=False,
                            scaling_bounds=False,
                            shearing_bounds=False,
                            translation_bounds=False,
                            enable_90_rotations=False):
    """build batchsize x 4 x 4 tensor representing an affine transformation in homogeneous coordinates.
    If return_inv is True, also returns the inverse of the created affine matrix."""

    if (rotation_bounds is not False) | (enable_90_rotations is not False):
        if n_dims == 2:
            if rotation_bounds is not False:
                rotation = draw_value_from_distribution(rotation_bounds,
                                                        size=1,
                                                        default_range=15.0,
                                                        return_as_tensor=True,
                                                        batchsize=batchsize)
            else:
                rotation = tf.zeros(tf.concat([batchsize, tf.ones(1, dtype='int32')], axis=0))
        else:  # n_dims = 3
            if rotation_bounds is not False:
                rotation = draw_value_from_distribution(rotation_bounds,
                                                        size=n_dims,
                                                        default_range=15.0,
                                                        return_as_tensor=True,
                                                        batchsize=batchsize)
            else:
                rotation = tf.zeros(tf.concat([batchsize, 3 * tf.ones(1, dtype='int32')], axis=0))
        if enable_90_rotations:
            rotation = tf.cast(tf.random.uniform(tf.shape(rotation), maxval=4, dtype='int32') * 90, 'float32') \
                       + rotation
        T_rot = create_rotation_transform(rotation, n_dims)
    else:
        T_rot = tf.tile(tf.expand_dims(tf.eye(n_dims), axis=0),
                        tf.concat([batchsize, tf.ones(2, dtype='int32')], axis=0))

    if shearing_bounds is not False:
        shearing = draw_value_from_distribution(shearing_bounds,
                                                size=n_dims ** 2 - n_dims,
                                                default_range=.01,
                                                return_as_tensor=True,
                                                batchsize=batchsize)
        T_shearing = create_shearing_transform(shearing, n_dims)
    else:
        T_shearing = tf.tile(tf.expand_dims(tf.eye(n_dims), axis=0),
                             tf.concat([batchsize, tf.ones(2, dtype='int32')], axis=0))

    if scaling_bounds is not False:
        scaling = draw_value_from_distribution(scaling_bounds,
                                               size=n_dims,
                                               centre=1,
                                               default_range=.15,
                                               return_as_tensor=True,
                                               batchsize=batchsize)
        T_scaling = tf.linalg.diag(scaling)
    else:
        T_scaling = tf.tile(tf.expand_dims(tf.eye(n_dims), axis=0),
                            tf.concat([batchsize, tf.ones(2, dtype='int32')], axis=0))

    T = tf.matmul(T_scaling, tf.matmul(T_shearing, T_rot))

    if translation_bounds is not False:
        translation = draw_value_from_distribution(translation_bounds,
                                                   size=n_dims,
                                                   default_range=5,
                                                   return_as_tensor=True,
                                                   batchsize=batchsize)
        T = tf.concat([T, tf.expand_dims(translation, axis=-1)], axis=-1)
    else:
        T = tf.concat([T, tf.zeros(tf.concat([tf.shape(T)[:2], tf.ones(1, dtype='int32')], 0))], axis=-1)

    # build rigid transform
    T_last_row = tf.expand_dims(tf.concat([tf.zeros((1, n_dims)), tf.ones((1, 1))], axis=1), 0)
    T_last_row = tf.tile(T_last_row, tf.concat([batchsize, tf.ones(2, dtype='int32')], axis=0))
    T = tf.concat([T, T_last_row], axis=1)

    return T


def create_rotation_transform(rotation, n_dims):
    """build rotation transform from 3d or 2d rotation coefficients. Angles are given in degrees."""
    rotation = rotation * np.pi / 180
    if n_dims == 3:
        shape = tf.shape(tf.expand_dims(rotation[..., 0], -1))

        Rx_row0 = tf.expand_dims(tf.tile(tf.expand_dims(tf.convert_to_tensor([1., 0., 0.]), 0), shape), axis=1)
        Rx_row1 = tf.stack([tf.zeros(shape), tf.expand_dims(tf.cos(rotation[..., 0]), -1),
                            tf.expand_dims(-tf.sin(rotation[..., 0]), -1)], axis=-1)
        Rx_row2 = tf.stack([tf.zeros(shape), tf.expand_dims(tf.sin(rotation[..., 0]), -1),
                            tf.expand_dims(tf.cos(rotation[..., 0]), -1)], axis=-1)
        Rx = tf.concat([Rx_row0, Rx_row1, Rx_row2], axis=1)

        Ry_row0 = tf.stack([tf.expand_dims(tf.cos(rotation[..., 1]), -1), tf.zeros(shape),
                            tf.expand_dims(tf.sin(rotation[..., 1]), -1)], axis=-1)
        Ry_row1 = tf.expand_dims(tf.tile(tf.expand_dims(tf.convert_to_tensor([0., 1., 0.]), 0), shape), axis=1)
        Ry_row2 = tf.stack([tf.expand_dims(-tf.sin(rotation[..., 1]), -1), tf.zeros(shape),
                            tf.expand_dims(tf.cos(rotation[..., 1]), -1)], axis=-1)
        Ry = tf.concat([Ry_row0, Ry_row1, Ry_row2], axis=1)

        Rz_row0 = tf.stack([tf.expand_dims(tf.cos(rotation[..., 2]), -1),
                            tf.expand_dims(-tf.sin(rotation[..., 2]), -1), tf.zeros(shape)], axis=-1)
        Rz_row1 = tf.stack([tf.expand_dims(tf.sin(rotation[..., 2]), -1),
                            tf.expand_dims(tf.cos(rotation[..., 2]), -1), tf.zeros(shape)], axis=-1)
        Rz_row2 = tf.expand_dims(tf.tile(tf.expand_dims(tf.convert_to_tensor([0., 0., 1.]), 0), shape), axis=1)
        Rz = tf.concat([Rz_row0, Rz_row1, Rz_row2], axis=1)

        T_rot = tf.matmul(tf.matmul(Rx, Ry), Rz)

    elif n_dims == 2:
        R_row0 = tf.stack([tf.expand_dims(tf.cos(rotation[..., 0]), -1),
                           tf.expand_dims(tf.sin(rotation[..., 0]), -1)], axis=-1)
        R_row1 = tf.stack([tf.expand_dims(-tf.sin(rotation[..., 0]), -1),
                           tf.expand_dims(tf.cos(rotation[..., 0]), -1)], axis=-1)
        T_rot = tf.concat([R_row0, R_row1], axis=1)

    else:
        raise Exception('only supports 2 or 3D.')

    return T_rot


def create_shearing_transform(shearing, n_dims):
    """build shearing transform from 2d/3d shearing coefficients"""
    shape = tf.shape(tf.expand_dims(shearing[..., 0], -1))
    if n_dims == 3:
        shearing_row0 = tf.stack([tf.ones(shape), tf.expand_dims(shearing[..., 0], -1),
                                  tf.expand_dims(shearing[..., 1], -1)], axis=-1)
        shearing_row1 = tf.stack([tf.expand_dims(shearing[..., 2], -1), tf.ones(shape),
                                  tf.expand_dims(shearing[..., 3], -1)], axis=-1)
        shearing_row2 = tf.stack([tf.expand_dims(shearing[..., 4], -1), tf.expand_dims(shearing[..., 5], -1),
                                  tf.ones(shape)], axis=-1)
        T_shearing = tf.concat([shearing_row0, shearing_row1, shearing_row2], axis=1)

    elif n_dims == 2:
        shearing_row0 = tf.stack([tf.ones(shape), tf.expand_dims(shearing[..., 0], -1)], axis=-1)
        shearing_row1 = tf.stack([tf.expand_dims(shearing[..., 1], -1), tf.ones(shape)], axis=-1)
        T_shearing = tf.concat([shearing_row0, shearing_row1], axis=1)
    else:
        raise Exception('only supports 2 or 3D.')
    return T_shearing


# --------------------------------------------------- miscellaneous ----------------------------------------------------


def infer(x):
    """ Try to parse input to float. If it fails, tries boolean, and otherwise keep it as string """
    try:
        x = float(x)
    except ValueError:
        if x == 'False':
            x = False
        elif x == 'True':
            x = True
        elif not isinstance(x, str):
            raise TypeError('input should be an int/float/boolean/str, had {}'.format(type(x)))
    return x


class LoopInfo:
    """
    Class to print the current iteration in a for loop, and optionally the estimated remaining time.
    Instantiate just before the loop, and call the update method at the start of the loop.
    The printed text has the following format:
    processing i/total    remaining time: hh:mm:ss
    """

    def __init__(self, n_iterations, spacing=10, text='processing', print_time=False):
        """
        :param n_iterations: total number of iterations of the for loop.
        :param spacing: frequency at which the update info will be printed on screen.
        :param text: text to print. Default is processing.
        :param print_time: whether to print the estimated remaining time. Default is False.
        """

        # loop parameters
        self.n_iterations = n_iterations
        self.spacing = spacing

        # text parameters
        self.text = text
        self.print_time = print_time
        self.print_previous_time = False
        self.align = len(str(self.n_iterations)) * 2 + 1 + 3

        # timing parameters
        self.iteration_durations = np.zeros((n_iterations,))
        self.start = time.time()
        self.previous = time.time()

    def update(self, idx):

        # time iteration
        now = time.time()
        self.iteration_durations[idx] = now - self.previous
        self.previous = now

        # print text
        if idx == 0:
            print(self.text + ' 1/{}'.format(self.n_iterations))
        elif idx % self.spacing == self.spacing - 1:
            iteration = str(idx + 1) + '/' + str(self.n_iterations)
            if self.print_time:
                # estimate remaining time
                max_duration = np.max(self.iteration_durations)
                average_duration = np.mean(self.iteration_durations[self.iteration_durations > .01 * max_duration])
                remaining_time = int(average_duration * (self.n_iterations - idx))
                # print total remaining time only if it is greater than 1s or if it was previously printed
                if (remaining_time > 1) | self.print_previous_time:
                    eta = str(timedelta(seconds=remaining_time))
                    print(self.text + ' {:<{x}} remaining time: {}'.format(iteration, eta, x=self.align))
                    self.print_previous_time = True
                else:
                    print(self.text + ' {}'.format(iteration))
            else:
                print(self.text + ' {}'.format(iteration))


def get_mapping_lut(source, dest=None):
    """This functions returns the look-up table to map a list of N values (source) to another list (dest).
    If the second list is not given, we assume it is equal to [0, ..., N-1]."""

    # initialise
    source = np.array(reformat_to_list(source), dtype='int32')
    n_labels = source.shape[0]

    # build new label list if necessary
    if dest is None:
        dest = np.arange(n_labels, dtype='int32')
    else:
        assert len(source) == len(dest), 'label_list and new_label_list should have the same length'
        dest = np.array(reformat_to_list(dest, dtype='int'))

    # build look-up table
    lut = np.zeros(np.max(source) + 1, dtype='int32')
    for source, dest in zip(source, dest):
        lut[source] = dest

    return lut


def build_training_generator(gen, batchsize):
    """Build generator for training a network."""
    while True:
        inputs = next(gen)
        if batchsize > 1:
            target = np.concatenate([np.zeros((1, 1))] * batchsize, 0)
        else:
            target = np.zeros((1, 1))
        yield inputs, target


def find_closest_number_divisible_by_m(n, m, answer_type='lower'):
    """Return the closest integer to n that is divisible by m. answer_type can either be 'closer', 'lower' (only returns
    values lower than n), or 'higher' (only returns values higher than m)."""
    if n % m == 0:
        return n
    else:
        q = int(n / m)
        lower = q * m
        higher = (q + 1) * m
        if answer_type == 'lower':
            return lower
        elif answer_type == 'higher':
            return higher
        elif answer_type == 'closer':
            return lower if (n - lower) < (higher - n) else higher
        else:
            raise Exception('answer_type should be lower, higher, or closer, had : %s' % answer_type)


def build_binary_structure(connectivity, n_dims, shape=None):
    """Return a dilation/erosion element with provided connectivity"""
    if shape is None:
        shape = [connectivity * 2 + 1] * n_dims
    else:
        shape = reformat_to_list(shape, length=n_dims)
    dist = np.ones(shape)
    center = tuple([tuple([int(s / 2)]) for s in shape])
    dist[center] = 0
    dist = distance_transform_edt(dist)
    struct = (dist <= connectivity) * 1
    return struct


def draw_value_from_distribution(hyperparameter,
                                 size=1,
                                 distribution='uniform',
                                 centre=0.,
                                 default_range=10.0,
                                 positive_only=False,
                                 return_as_tensor=False,
                                 batchsize=None):
    """Sample values from a uniform, or normal distribution of given hyperparameters.
    These hyperparameters are to the number of 2 in both uniform and normal cases.
    :param hyperparameter: values of the hyperparameters. Can either be:
    1) None, in each case the two hyperparameters are given by [center-default_range, center+default_range],
    2) a number, where the two hyperparameters are given by [centre-hyperparameter, centre+hyperparameter],
    3) a sequence of length 2, directly defining the two hyperparameters: [min, max] if the distribution is uniform,
    [mean, std] if the distribution is normal.
    4) a numpy array, with size (2, m). In this case, the function returns a 1d array of size m, where each value has
    been sampled independently with the specified hyperparameters. If the distribution is uniform, rows correspond to
    its lower and upper bounds, and if the distribution is normal, rows correspond to its mean and std deviation.
    5) a numpy array of size (2*n, m). Same as 4) but we first randomly select a block of two rows among the
    n possibilities.
    6) the path to a numpy array corresponding to case 4 or 5.
    7) False, in which case this function returns None.
    :param size: (optional) number of values to sample. All values are sampled independently.
    Used only if hyperparameter is not a numpy array.
    :param distribution: (optional) the distribution type. Can be 'uniform' or 'normal'. Default is 'uniform'.
    :param centre: (optional) default centre to use if hyperparameter is None or a number.
    :param default_range: (optional) default range to use if hyperparameter is None.
    :param positive_only: (optional) whether to reset all negative values to zero.
    :param return_as_tensor: (optional) whether to return the result as a tensorflow tensor
    :param batchsize: (optional) if return_as_tensor is true, then you can sample a tensor of a given batchsize. Give
    this batchsize as a tensorflow tensor here.
    :return: a float, or a numpy 1d array if size > 1, or hyperparameter is itself a numpy array.
    Returns None if hyperparameter is False.
    """

    # return False is hyperparameter is False
    if hyperparameter is False:
        return None

    # reformat parameter_range
    hyperparameter = load_array_if_path(hyperparameter, load_as_numpy=True)
    if not isinstance(hyperparameter, np.ndarray):
        if hyperparameter is None:
            hyperparameter = np.array([[centre - default_range] * size, [centre + default_range] * size])
        elif isinstance(hyperparameter, (int, float)):
            hyperparameter = np.array([[centre - hyperparameter] * size, [centre + hyperparameter] * size])
        elif isinstance(hyperparameter, (list, tuple)):
            assert len(hyperparameter) == 2, 'if list, parameter_range should be of length 2.'
            hyperparameter = np.transpose(np.tile(np.array(hyperparameter), (size, 1)))
        else:
            raise ValueError('parameter_range should either be None, a number, a sequence, or a numpy array.')
    elif isinstance(hyperparameter, np.ndarray):
        assert hyperparameter.shape[0] % 2 == 0, 'number of rows of parameter_range should be divisible by 2'
        n_modalities = int(hyperparameter.shape[0] / 2)
        modality_idx = 2 * np.random.randint(n_modalities)
        hyperparameter = hyperparameter[modality_idx: modality_idx + 2, :]

    # draw values as tensor
    if return_as_tensor:
        shape = KL.Lambda(lambda x: tf.convert_to_tensor(hyperparameter.shape[1], 'int32'))([])
        if batchsize is not None:
            shape = KL.Lambda(lambda x: tf.concat([x[0], tf.expand_dims(x[1], axis=0)], axis=0))([batchsize, shape])
        if distribution == 'uniform':
            parameter_value = KL.Lambda(lambda x: tf.random.uniform(shape=x,
                                                                    minval=hyperparameter[0, :],
                                                                    maxval=hyperparameter[1, :]))(shape)
        elif distribution == 'normal':
            parameter_value = KL.Lambda(lambda x: tf.random.normal(shape=x,
                                                                   mean=hyperparameter[0, :],
                                                                   stddev=hyperparameter[1, :]))(shape)
        else:
            raise ValueError("Distribution not supported, should be 'uniform' or 'normal'.")

        if positive_only:
            parameter_value = KL.Lambda(lambda x: K.clip(x, 0, None))(parameter_value)

    # draw values as numpy array
    else:
        if distribution == 'uniform':
            parameter_value = np.random.uniform(low=hyperparameter[0, :], high=hyperparameter[1, :])
        elif distribution == 'normal':
            parameter_value = np.random.normal(loc=hyperparameter[0, :], scale=hyperparameter[1, :])
        else:
            raise ValueError("Distribution not supported, should be 'uniform' or 'normal'.")

        if positive_only:
            parameter_value[parameter_value < 0] = 0

    return parameter_value


def build_exp(x, first, last, fix_point):
    # first = f(0), last = f(+inf), fix_point = [x0, f(x0))]
    a = last
    b = first - last
    c = - (1 / fix_point[0]) * np.log((fix_point[1] - last) / (first - last))
    return a + b * np.exp(-c * x)



"""
apply_warp - Image registration transformation application

Part of the LAMAR processing pipeline for neuroimaging data.

This module applies spatial transformations to register images from one space to another
using both affine and non-linear (warp field) transformations. It's commonly used to:
- Transform subject images to a standard space (e.g., MNI152)
- Register images across modalities (e.g., T1w to FLAIR)
- Apply previously calculated transformations to derived images (e.g., segmentations)

The module leverages ANTsPy to apply the transformations in the correct order.

Transform Order:
---------------
By default (inverse=False), transforms are applied in this order:
    moving → [secondary_warp] → [warp] → [affine] → fixed

When inverse=True, the order is reversed:
    moving → [affine^-1] → [warp] → [secondary_warp] → fixed

Note: In ANTs' transformlist, the LAST transform in the list is applied FIRST.
Therefore, to achieve moving → warp → affine, we provide [warp, affine] to ANTs.

Robust Mode (Two Warpfields):
-----------------------------
When using LAMAR's robust registration mode, you'll have two warpfields:
1. Primary warpfield: From parcellation-based registration (coarse alignment)
2. Secondary warpfield: From direct image registration (fine-tuning)

To apply both, use:
    --warp primary_warp.nii.gz --secondary-warp secondary_warp.nii.gz

The composition order is: moving → secondary → primary → affine → fixed

API Usage:
---------
lamar apply-warp \\
    --moving <path/to/source_image.nii.gz> \\
    --fixed <path/to/target_space.nii.gz> \\
    --affine <path/to/transform.mat> \\
    --warp <path/to/warpfield.nii.gz> \\
    [--secondary-warp <path/to/secondary_warpfield.nii.gz>] \\
    [--output <path/to/registered_image.nii.gz>] \\
    [--interpolation linear|nearestNeighbor|multiLabel] \\
    [--inverse]

Python Usage:
-----------
>>> import ants
>>> from lamareg.scripts.apply_warp import apply_warp
>>> moving_img = ants.image_read("subject_t1w.nii.gz")
>>> reference_img = ants.image_read("mni152.nii.gz")
>>> apply_warp(
...     moving_img=moving_img,
...     reference_img=reference_img,
...     affine_file="transform.mat",
...     warp_file="warpfield.nii.gz",
...     secondary_warp="secondary_warp.nii.gz",
...     out_file="registered_t1w.nii.gz",
...     interpolation="linear"
... )

Examples:
--------
# Apply single warpfield + affine (single-stage registration):
lamar apply-warp \\
    --moving subject_flair.nii.gz \\
    --fixed subject_t1w.nii.gz \\
    --warp flair_to_t1w_warp.nii.gz \\
    --affine flair_to_t1w_affine.mat \\
    --output registered_flair.nii.gz

# Apply two warpfields + affine (robust mode):
lamar apply-warp \\
    --moving dwi_segmentation.nii.gz \\
    --fixed T1w_reference.nii.gz \\
    --warp dwi_to_T1w_primary_warp.nii.gz \\
    --secondary-warp dwi_to_T1w_secondary_warp.nii.gz \\
    --affine dwi_to_T1w_affine.mat \\
    --output dwi_seg_in_T1w.nii.gz

# Apply inverse transforms (from fixed back to moving space):
lamar apply-warp \\
    --moving atlas_in_T1w.nii.gz \\
    --fixed subject_dwi.nii.gz \\
    --warp T1w_to_dwi_warp.nii.gz \\
    --affine T1w_to_dwi_affine.mat \\
    --output atlas_in_dwi.nii.gz \\
    --inverse

# Apply to label image (use nearest neighbor interpolation):
lamar apply-warp \\
    --moving segmentation.nii.gz \\
    --fixed mni152.nii.gz \\
    --warp warp.nii.gz \\
    --affine affine.mat \\
    --output seg_in_mni.nii.gz \\
    --interpolation nearestNeighbor

References:
----------
1. Avants BB, Tustison NJ, Song G, et al. A reproducible evaluation of ANTs
   similarity metric performance in brain image registration. NeuroImage.
   2011;54(3):2033-2044. doi:10.1016/j.neuroimage.2010.09.025
"""

import ants
import argparse
import sys
from colorama import init, Fore, Style

init()


def print_help():
    """Print a help message with examples."""
    # ANSI color codes
    CYAN = Fore.CYAN
    GREEN = Fore.GREEN
    YELLOW = Fore.YELLOW
    BLUE = Fore.BLUE
    MAGENTA = Fore.MAGENTA
    BOLD = Style.BRIGHT
    RESET = Style.RESET_ALL

    help_text = f"""
    {CYAN}{BOLD}╔════════════════════════════════════════════════════════════════╗
    ║                        APPLY WARP                              ║
    ╚════════════════════════════════════════════════════════════════╝{RESET}
    
    This script applies spatial transformations (affine + warp fields) to
    register a moving image to a reference space.
    
    {CYAN}{BOLD}────────────────────────── REQUIRED ARGUMENTS ──────────────────────────{RESET}
      {YELLOW}--moving{RESET}     : Path to the input image to be warped (.nii.gz)
      {YELLOW}--fixed{RESET}      : Path to the target/reference image (.nii.gz)
    
    {CYAN}{BOLD}────────────────────────── OPTIONAL ARGUMENTS ──────────────────────────{RESET}
      {YELLOW}--affine{RESET}            : Path to the affine transformation file (.mat)
      {YELLOW}--warp{RESET}              : Path to the primary warp field (.nii.gz)
      {YELLOW}--secondary-warp{RESET}    : Path to the secondary warp field (.nii.gz) for robust mode
      {YELLOW}--output{RESET}            : Output path for the warped image (default: warped_image.nii.gz)
      {YELLOW}--interpolation{RESET}     : Interpolation method (default: linear)
                             Options: linear, nearestNeighbor, multiLabel, bSpline
      {YELLOW}--inverse{RESET}           : Apply inverse transforms (fixed → moving direction)
    
    {CYAN}{BOLD}────────────────────────── TRANSFORM ORDER ──────────────────────────{RESET}
    
    {BLUE}Forward (default):{RESET}
      moving → {MAGENTA}[secondary_warp]{RESET} → {MAGENTA}[warp]{RESET} → {MAGENTA}[affine]{RESET} → fixed
    
    {BLUE}Inverse (--inverse flag):{RESET}
      fixed → {MAGENTA}[affine^-1]{RESET} → {MAGENTA}[warp]{RESET} → {MAGENTA}[secondary_warp]{RESET} → moving
    
    {CYAN}{BOLD}────────────────────────── EXAMPLE USAGE ──────────────────────────{RESET}
    
    {BLUE}# Single-stage registration (warp + affine):{RESET}
    lamar {GREEN}apply-warp{RESET} {YELLOW}--moving{RESET} subject_flair.nii.gz {YELLOW}--fixed{RESET} subject_t1w.nii.gz \\
      {YELLOW}--warp{RESET} flair_to_t1w_warp.nii.gz {YELLOW}--affine{RESET} flair_to_t1w_affine.mat \\
      {YELLOW}--output{RESET} registered_flair.nii.gz
    
    {BLUE}# Robust mode (two warpfields + affine):{RESET}
    lamar {GREEN}apply-warp{RESET} {YELLOW}--moving{RESET} dwi.nii.gz {YELLOW}--fixed{RESET} T1w.nii.gz \\
      {YELLOW}--warp{RESET} dwi_to_T1w_primary.nii.gz \\
      {YELLOW}--secondary-warp{RESET} dwi_to_T1w_secondary.nii.gz \\
      {YELLOW}--affine{RESET} dwi_to_T1w_affine.mat \\
      {YELLOW}--output{RESET} dwi_in_T1w.nii.gz
    
    {BLUE}# Apply to segmentation (nearest neighbor interpolation):{RESET}
    lamar {GREEN}apply-warp{RESET} {YELLOW}--moving{RESET} parcellation.nii.gz {YELLOW}--fixed{RESET} mni152.nii.gz \\
      {YELLOW}--warp{RESET} warp.nii.gz {YELLOW}--affine{RESET} affine.mat \\
      {YELLOW}--interpolation{RESET} nearestNeighbor \\
      {YELLOW}--output{RESET} parc_in_mni.nii.gz
    
    {BLUE}# Apply inverse transforms:{RESET}
    lamar {GREEN}apply-warp{RESET} {YELLOW}--moving{RESET} atlas_in_T1w.nii.gz {YELLOW}--fixed{RESET} dwi.nii.gz \\
      {YELLOW}--warp{RESET} T1w_to_dwi_warp.nii.gz {YELLOW}--affine{RESET} T1w_to_dwi_affine.mat \\
      {YELLOW}--inverse{RESET} {YELLOW}--output{RESET} atlas_in_dwi.nii.gz
    
    {CYAN}{BOLD}────────────────────────── NOTES ───────────────────────────{RESET}
    
    {MAGENTA}•{RESET} At least one transform (--affine, --warp, or --secondary-warp) must be provided
    {MAGENTA}•{RESET} For label/segmentation images, use --interpolation nearestNeighbor or multiLabel
    {MAGENTA}•{RESET} For continuous images (T1w, FLAIR, etc.), use --interpolation linear or bSpline
    {MAGENTA}•{RESET} The --inverse flag inverts the affine but applies warpfields as-is
    {MAGENTA}•{RESET} Robust mode requires both --warp and --secondary-warp from LAMAR registration
    """

    print(help_text)


def apply_warp(
    moving_img, reference_img, affine_file, warp_file, out_file, interpolation="linear", inverse=False, secondary_warp=None
):
    """Apply affine and warp field transformations to a moving image.

    This function takes a moving image and applies spatial transformations to register
    it to a reference image space. It supports both single-stage registration (one warp)
    and robust two-stage registration (primary + secondary warp).

    Parameters
    ----------
    moving_img : ants.ANTsImage
        The moving image that will be transformed.
    reference_img : ants.ANTsImage
        The reference/fixed image that defines the target space.
    affine_file : str or None
        Path to the affine transformation file (.mat). Can be None if not needed.
    warp_file : str or None
        Path to the primary nonlinear warp field (.nii.gz). Can be None if not needed.
    out_file : str
        Path where the transformed image will be saved.
    interpolation : str, optional
        Interpolation method to use for the transformation. Default is 'linear'.
        Options: 'linear', 'nearestNeighbor', 'multiLabel', 'bSpline', 'gaussian', etc.
    inverse : bool, optional
        If True, reverses the transform order and inverts the affine. Default is False.
    secondary_warp : str or None, optional
        Path to the secondary warp field (.nii.gz) for robust mode. Default is None.

    Returns
    -------
    None
        The function saves the transformed image to the specified output path.

    Notes
    -----
    Transform Order (inverse=False):
        moving → [secondary_warp] → [warp] → [affine] → fixed
    
    Transform Order (inverse=True):
        moving → [affine^-1] → [warp] → [secondary_warp] → fixed
    
    ANTs applies transforms in reverse order of the transformlist, so the last
    transform in the list is applied first. This function handles the ordering
    automatically based on the inverse flag.
    
    For robust mode (two-stage registration), both warp_file and secondary_warp
    should be provided. The secondary warp is applied first, followed by the
    primary warp, then the affine.

    Examples
    --------
    >>> import ants
    >>> from lamareg.scripts.apply_warp import apply_warp
    >>> moving = ants.image_read("dwi.nii.gz")
    >>> fixed = ants.image_read("T1w.nii.gz")
    >>> 
    >>> # Single-stage registration
    >>> apply_warp(moving, fixed, "affine.mat", "warp.nii.gz", "output.nii.gz")
    >>> 
    >>> # Robust mode (two warpfields)
    >>> apply_warp(moving, fixed, "affine.mat", "primary.nii.gz", "output.nii.gz",
    ...            secondary_warp="secondary.nii.gz")
    >>> 
    >>> # Apply to segmentation
    >>> apply_warp(moving, fixed, "affine.mat", "warp.nii.gz", "seg_out.nii.gz",
    ...            interpolation="nearestNeighbor")
    """

    # The order of transforms in transformlist matters (last transform will be applied first).
    # Usually you put the nonlinear warp first, then the affine:
    if not (affine_file or warp_file or secondary_warp):
        print("ERROR: At least one of --affine, --warp, or --secondary-warp must be provided.")
        sys.exit(1)
    
    if inverse:
        # Inverse order: affine^-1, then warp, then secondary_warp
        transformlist = []
        invertlist = []
        if affine_file:
            transformlist.append(affine_file)
            invertlist.append(True)  # Invert the affine
        if warp_file:
            transformlist.append(warp_file)
            invertlist.append(False)  # Don't invert warp (already in inverse direction)
        if secondary_warp:
            transformlist.append(secondary_warp)
            invertlist.append(False)  # Don't invert secondary warp

        transformed = ants.apply_transforms(
            fixed=reference_img,
            moving=moving_img,
            transformlist=transformlist,
            interpolator=interpolation,
            whichtoinvert=invertlist
        )
    else:
        # Forward order: secondary_warp, then warp, then affine
        # ANTs applies in reverse, so list them backwards
        transformlist = []
        if secondary_warp:
            transformlist.append(secondary_warp)
        if warp_file:
            transformlist.append(warp_file)
        if affine_file:
            transformlist.append(affine_file)
        
        transformed = ants.apply_transforms(
            fixed=reference_img,
            moving=moving_img,
            transformlist=transformlist,
            interpolator=interpolation,
        )

    # Save the transformed image
    ants.image_write(transformed, out_file)
    print(f"Saved warped image as {out_file}")


def main():
    parser = argparse.ArgumentParser(
        description="Apply affine and warp field transformations to an image using ANTsPy.",
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument(
        "--moving", required=True, help="Path to the moving image (.nii.gz)."
    )
    parser.add_argument(
        "--fixed", required=True, help="Path to the fixed/reference image (.nii.gz)."
    )
    parser.add_argument(
        "--output", default="warped_image.nii.gz", help="Output warped image filename."
    )
    parser.add_argument(
        "--affine", help="Path to the affine transform (.mat)."
    )
    parser.add_argument(
        "--warp", help="Path to the primary warp field (.nii.gz)."
    )
    parser.add_argument(
        "--secondary-warp", help="Path to the secondary warp field (.nii.gz) for robust mode."
    )
    parser.add_argument(
        "--interpolation",
        default="linear",
        help="Interpolation method (default: linear). Options: linear, nearestNeighbor, multiLabel, bSpline."
    )
    parser.add_argument(
        "--inverse",
        action="store_true",
        help="Apply inverse transforms (fixed → moving direction, inverts affine)"
    )
    args = parser.parse_args()

    moving_img = ants.image_read(args.moving)
    reference_img = ants.image_read(args.fixed)

    apply_warp(
        moving_img,
        reference_img,
        args.affine,
        args.warp,
        args.output,
        args.interpolation,
        inverse=args.inverse,
        secondary_warp=args.secondary_warp if hasattr(args, 'secondary_warp') else None
    )


if __name__ == "__main__":
    main()



"""
synthseg - Neural Network-Based Brain MRI Segmentation

Part of the micaflow processing pipeline for neuroimaging data.

This module provides an interface to SynthSeg, a deep learning-based tool for automated
brain MRI segmentation that works across different MRI contrasts without retraining.
SynthSeg segments brain anatomical structures in T1w, T2w, FLAIR, and other MR contrasts,
providing both whole-brain segmentation and optional cortical parcellation.

Features:
--------
- Contrast-agnostic segmentation working across different MRI acquisition types
- Whole-brain anatomical structure segmentation with 37 labels
- Optional cortical parcellation (up to 95 additional regions)
- Multiple execution modes: standard, robust (higher quality), and fast
- Volumetric analysis with CSV output for region-wise measurements
- Quality control metrics for assessing segmentation reliability
- GPU acceleration with optional CPU-only execution

API Usage:
---------
micaflow synthseg
    --i <path/to/image.nii.gz>
    --o <path/to/segmentation.nii.gz>
    [--parc]
    [--robust]
    [--fast]
    [--vol <path/to/volumes.csv>]
    [--qc <path/to/qc_scores.csv>]
    [--threads <num_threads>]

Python Usage:
-----------
>>> from micaflow.scripts.synthseg import main
>>> main({
...     'i': 'input_image.nii.gz',
...     'o': 'segmentation.nii.gz',
...     'parc': True,
...     'robust': False,
...     'fast': True,
...     'vol': 'volumes.csv',
...     'threads': 4
... })

"""

# python imports
import os
import sys
import multiprocessing
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from colorama import init, Fore, Style

# Get number of available CPU cores
DEFAULT_THREADS = multiprocessing.cpu_count()

init()


def print_extended_help():
    """Print extended help message with examples and usage instructions."""
    # ANSI color codes
    CYAN = Fore.CYAN
    GREEN = Fore.GREEN
    YELLOW = Fore.YELLOW
    BLUE = Fore.BLUE
    MAGENTA = Fore.MAGENTA
    BOLD = Style.BRIGHT
    RESET = Style.RESET_ALL

    help_text = f"""
    {CYAN}{BOLD}╔════════════════════════════════════════════════════════════════╗
    ║                         SYNTHSEG                               ║
    ╚════════════════════════════════════════════════════════════════╝{RESET}
    
    This script runs the SynthSeg neural network-based tool for brain MRI
    segmentation. It provides automated segmentation of anatomical structures
    even across different contrasts and acquisition types.
    
    {CYAN}{BOLD}────────────────────────── USAGE ──────────────────────────{RESET}
      micaflow synthseg {GREEN}[options]{RESET}
    
    {CYAN}{BOLD}─────────────────── REQUIRED ARGUMENTS ───────────────────{RESET}
      {YELLOW}--i{RESET} PATH       : Input image(s) to segment (file or folder)
      {YELLOW}--o{RESET} PATH       : Output segmentation file(s) or folder
    
    {CYAN}{BOLD}─────────────────── OPTIONAL ARGUMENTS ───────────────────{RESET}
      {YELLOW}--parc{RESET}         : Enable cortical parcellation
      {YELLOW}--robust{RESET}       : Use robust mode (slower but better quality)
      {YELLOW}--fast{RESET}         : Faster processing (less postprocessing)
      {YELLOW}--threads{RESET} N    : Set number of CPU threads (default: 1)
      {YELLOW}--cpu{RESET}          : Force CPU processing (instead of GPU)
      {YELLOW}--vol{RESET} PATH     : Output volumetric CSV file
      {YELLOW}--qc{RESET} PATH      : Output quality control scores CSV file
      {YELLOW}--post{RESET} PATH    : Output posterior probability maps
      {YELLOW}--resample{RESET} PATH: Output resampled images
      {YELLOW}--crop{RESET} N [N ...]: Size of 3D patches to analyze (default: 192)
      {YELLOW}--ct{RESET}           : Clip intensities for CT scans [0,80]
      {YELLOW}--v1{RESET}           : Use SynthSeg 1.0 instead of 2.0
    
    {CYAN}{BOLD}────────────────── EXAMPLE USAGE ────────────────────────{RESET}
    
    {BLUE}# Basic segmentation{RESET}
    micaflow synthseg \\
      {YELLOW}--i{RESET} t1w_scan.nii.gz \\
      {YELLOW}--o{RESET} segmentation.nii.gz
    
    {BLUE}# With cortical parcellation{RESET}
    micaflow synthseg \\
      {YELLOW}--i{RESET} t1w_scan.nii.gz \\
      {YELLOW}--o{RESET} segmentation.nii.gz \\
      {YELLOW}--parc{RESET}
    
    {BLUE}# Batch processing with volume calculation{RESET}
    micaflow synthseg \\
      {YELLOW}--i{RESET} input_folder/ \\
      {YELLOW}--o{RESET} output_folder/ \\
      {YELLOW}--vol{RESET} volumes.csv
    
    {CYAN}{BOLD}────────────────────────── NOTES ───────────────────────{RESET}
    {MAGENTA}•{RESET} SynthSeg works with any MRI contrast without retraining
    {MAGENTA}•{RESET} GPU acceleration is used by default for faster processing
    {MAGENTA}•{RESET} The robust mode provides better quality but is slower
    {MAGENTA}•{RESET} For batch processing, input and output paths must be folders
    """
    print(help_text)


def main(args):
    synthseg_home = os.path.dirname(os.path.abspath(__file__))
    sys.path.append(synthseg_home)
    model_dir = os.path.join(synthseg_home, "models")
    labels_dir = os.path.join(synthseg_home, "data/labels_classes_priors")
    # The rest of your code remains unchanged
    # print SynthSeg version and checks boolean params for SynthSeg-robust
    if args["robust"]:
        args["fast"] = True
        assert not args[
            "v1"
        ], "The flag --v1 cannot be used with --robust since SynthSeg-robust only came out with 2.0."
        version = "SynthSeg-robust 2.0"
    else:
        version = "SynthSeg 1.0" if args["v1"] else "SynthSeg 2.0"
        if args["fast"]:
            version += " (fast)"
    print("\n" + version + "\n")

    # enforce CPU processing if necessary
    if args["cpu"]:
        print("using CPU, hiding all CUDA_VISIBLE_DEVICES")
        os.environ["CUDA_VISIBLE_DEVICES"] = "-1"

    # limit the number of threads to be used if running on CPU
    import tensorflow as tf

    args["threads"] = int(args["threads"])
    if args["threads"] == 1:
        print("using 1 thread")
    else:
        print("using %s threads" % args["threads"])
    tf.config.threading.set_inter_op_parallelism_threads(args["threads"])
    tf.config.threading.set_intra_op_parallelism_threads(args["threads"])

    # path models
    if args["robust"]:
        args["path_model_segmentation"] = os.path.join(
            model_dir, "synthseg_robust_2.0.h5"
        )
    else:
        args["path_model_segmentation"] = os.path.join(model_dir, "synthseg_2.0.h5")
    args["path_model_parcellation"] = os.path.join(model_dir, "synthseg_parc_2.0.h5")
    args["path_model_qc"] = os.path.join(model_dir, "synthseg_qc_2.0.h5")

    # path labels
    args["labels_segmentation"] = os.path.join(
        labels_dir, "synthseg_segmentation_labels_2.0.npy"
    )
    args["labels_denoiser"] = os.path.join(
        labels_dir, "synthseg_denoiser_labels_2.0.npy"
    )
    args["labels_parcellation"] = os.path.join(
        labels_dir, "synthseg_parcellation_labels.npy"
    )
    args["labels_qc"] = os.path.join(labels_dir, "synthseg_qc_labels_2.0.npy")
    args["names_segmentation_labels"] = os.path.join(
        labels_dir, "synthseg_segmentation_names_2.0.npy"
    )
    args["names_parcellation_labels"] = os.path.join(
        labels_dir, "synthseg_parcellation_names.npy"
    )
    args["names_qc_labels"] = os.path.join(labels_dir, "synthseg_qc_names_2.0.npy")
    args["topology_classes"] = os.path.join(
        labels_dir, "synthseg_topological_classes_2.0.npy"
    )
    args["n_neutral_labels"] = 19

    # use previous model if needed
    if args["v1"]:
        args["path_model_segmentation"] = os.path.join(model_dir, "synthseg_1.0.h5")
        args["labels_segmentation"] = args["labels_segmentation"].replace(
            "_2.0.npy", ".npy"
        )
        args["labels_qc"] = args["labels_qc"].replace("_2.0.npy", ".npy")
        args["names_segmentation_labels"] = args["names_segmentation_labels"].replace(
            "_2.0.npy", ".npy"
        )
        args["names_qc_labels"] = args["names_qc_labels"].replace("_2.0.npy", ".npy")
        args["topology_classes"] = args["topology_classes"].replace("_2.0.npy", ".npy")
        args["n_neutral_labels"] = 18

    from lamareg.SynthSeg.predict_synthseg import predict

    # run prediction
    predict(
        path_images=args["i"],
        path_segmentations=args["o"],
        path_model_segmentation=args["path_model_segmentation"],
        labels_segmentation=args["labels_segmentation"],
        robust=args["robust"],
        fast=args["fast"],
        v1=args["v1"],
        do_parcellation=args["parc"],
        n_neutral_labels=args["n_neutral_labels"],
        names_segmentation=args["names_segmentation_labels"],
        labels_denoiser=args["labels_denoiser"],
        path_posteriors=args["post"],
        path_resampled=args["resample"],
        path_volumes=args["vol"],
        path_model_parcellation=args["path_model_parcellation"],
        labels_parcellation=args["labels_parcellation"],
        names_parcellation=args["names_parcellation_labels"],
        path_model_qc=args["path_model_qc"],
        labels_qc=args["labels_qc"],
        path_qc_scores=args["qc"],
        names_qc=args["names_qc_labels"],
        cropping=args["crop"],
        topology_classes=args["topology_classes"],
        ct=args["ct"],
    )


if __name__ == "__main__":
    # Check if help flags are provided or no arguments
    if len(sys.argv) == 1 or "-h" in sys.argv or "--help" in sys.argv:
        print_extended_help()
        sys.exit(0)

    # parse arguments
    parser = ArgumentParser(
        description="SynthSeg: Deep learning tool for brain MRI segmentation",
        epilog="For more details see: https://github.com/BBillot/SynthSeg",
        formatter_class=RawDescriptionHelpFormatter,
    )

    # input/outputs
    parser.add_argument(
        "--i", help="Image(s) to segment. Can be a path to an image or to a folder."
    )
    parser.add_argument(
        "--o",
        help="Segmentation output(s). Must be a folder if --i designates a folder.",
    )
    parser.add_argument(
        "--parc",
        action="store_true",
        help="(optional) Whether to perform cortex parcellation.",
    )
    parser.add_argument(
        "--robust",
        action="store_true",
        help="(optional) Whether to use robust predictions (slower).",
    )
    parser.add_argument(
        "--fast",
        action="store_true",
        help="(optional) Bypass some postprocessing for faster predictions.",
    )
    parser.add_argument(
        "--ct",
        action="store_true",
        help="(optional) Clip intensities to [0,80] for CT scans.",
    )
    parser.add_argument(
        "--vol",
        help="(optional) Path to output CSV file with volumes (mm3) for all regions and subjects.",
    )
    parser.add_argument(
        "--qc",
        help="(optional) Path to output CSV file with qc scores for all subjects.",
    )
    parser.add_argument(
        "--post",
        help="(optional) Posteriors output(s). Must be a folder if --i designates a folder.",
    )
    parser.add_argument(
        "--resample",
        help="(optional) Resampled image(s). Must be a folder if --i designates a folder.",
    )
    parser.add_argument(
        "--crop",
        nargs="+",
        type=int,
        help="(optional) Size of 3D patches to analyse. Default is 192.",
    )
    parser.add_argument(
        "--threads",
        type=int,
        default=DEFAULT_THREADS,
        help=f"(optional) Number of cores to be used. Default is {DEFAULT_THREADS} (all available).",
    )
    parser.add_argument(
        "--cpu",
        action="store_true",
        help="(optional) Enforce running with CPU rather than GPU.",
    )
    parser.add_argument(
        "--v1",
        action="store_true",
        help="(optional) Use SynthSeg 1.0 (updated 25/06/22).",
    )

    # parse commandline
    args = vars(parser.parse_args())
    main(args)



"""
 Dice Score Comparison for Brain Parcellation Maps

This script compares two brain parcellation images and calculates the Dice similarity coefficient for each label (region).

Features:
- Computes Dice score per anatomical label.
- Maps label numbers to region names using FreeSurfer and Desikan-Killiany label conventions.
- Outputs a readable CSV file with label, region name, and Dice score.
- Accepts command-line arguments using argparse:
    --ref : reference parcellation image (e.g., fixed image)
    --reg : registered parcellation image (e.g., moving image after registration)
    --out : output CSV file to save results

This script helps to evaluate the accuracy of image registration or segmentation
by comparing anatomical agreement between two labeled brain volumes.
""" 


import nibabel as nib
import numpy as np
from collections import defaultdict
import csv
import os
import argparse
import sys
from colorama import init, Fore, Style

init()

# FreeSurfer label-to-region mapping
FREESURFER_LABELS = {
    0: "Background", 2: "Left cerebral white matter", 3: "Left cerebral cortex",
    4: "Left lateral ventricle", 5: "Left inferior lateral ventricle",
    7: "Left cerebellum white matter", 8: "Left cerebellum cortex",
    10: "Left thalamus", 11: "Left caudate", 12: "Left putamen", 13: "Left pallidum",
    14: "3rd ventricle", 15: "4th ventricle", 16: "Brain-stem",
    17: "Left hippocampus", 18: "Left amygdala", 24: "CSF", 26: "Left accumbens area",
    28: "Left ventral DC", 41: "Right cerebral white matter", 42: "Right cerebral cortex",
    43: "Right lateral ventricle", 44: "Right inferior lateral ventricle",
    46: "Right cerebellum white matter", 47: "Right cerebellum cortex",
    49: "Right thalamus", 50: "Right caudate", 51: "Right putamen",
    52: "Right pallidum", 53: "Right hippocampus", 54: "Right amygdala",
    58: "Right accumbens area", 60: "Right ventral DC"
}

# Desikan-Killiany cortical labels (1001–1035 = left, 2001–2035 = right)
desikan_labels = [
    "banks STS", "caudal anterior cingulate", "caudal middle frontal", "corpuscallosum", "cuneus",
    "entorhinal", "fusiform", "inferior parietal", "inferior temporal",
    "isthmus cingulate", "lateral occipital", "lateral orbitofrontal",
    "lingual", "medial orbitofrontal", "middle temporal", "parahippocampal",
    "paracentral", "pars opercularis", "pars orbitalis", "pars triangularis",
    "pericalcarine", "postcentral", "posterior cingulate", "precentral",
    "precuneus", "rostral anterior cingulate", "rostral middle frontal",
    "superior frontal", "superior parietal", "superior temporal",
    "supramarginal", "frontal pole", "temporal pole", "transverse temporal", "insula"
]

for i, name in enumerate(desikan_labels):
    FREESURFER_LABELS[1001 + i] = f"Left {name}"
    FREESURFER_LABELS[2001 + i] = f"Right {name}"

def dice_score(mask1, mask2):
    intersection = np.sum((mask1 > 0) & (mask2 > 0))
    size1 = np.sum(mask1 > 0)
    size2 = np.sum(mask2 > 0)
    if size1 + size2 == 0:
        return np.nan
    return 2.0 * intersection / (size1 + size2)

def compare_parcellations_dice(parc1_path, parc2_path, output_csv_path):
    """
    Compare two parcellation images and calculate Dice scores for each label.
    
    Automatically resamples images to the same space if needed using nearest neighbor
    interpolation to preserve label values.
    """
    # Load the parcellation images with nibabel to check dimensions
    parc1_img = nib.load(parc1_path)
    parc2_img = nib.load(parc2_path)
    
    # Check if the images are in the same space
    same_shape = parc1_img.shape == parc2_img.shape
    same_affine = np.allclose(parc1_img.affine, parc2_img.affine, atol=1e-3)
    
    if not (same_shape and same_affine):
        print("\nWARNING: Images are not in the same space.")
        print(f"Image 1 shape: {parc1_img.shape}, Image 2 shape: {parc2_img.shape}")
        print("Resampling to the larger image's space using nearest neighbor interpolation...")
        
        # Determine which image has the larger dimensions (by volume)
        vol1 = np.prod(parc1_img.shape)
        vol2 = np.prod(parc2_img.shape)
        
        # Use ANTs for resampling with nearest neighbor interpolation
        import ants
        
        parc1_ants = ants.image_read(parc1_path)
        parc2_ants = ants.image_read(parc2_path)
        
        if vol1 >= vol2:
            # Resample parc2 to parc1's space
            print(f"Resampling {os.path.basename(parc2_path)} to match {os.path.basename(parc1_path)}")
            resampled = ants.resample_image_to_target(
                parc2_ants, parc1_ants, interp_type='nearestNeighbor')
            parc1 = parc1_ants.numpy()
            parc2 = resampled.numpy()
        else:
            # Resample parc1 to parc2's space
            print(f"Resampling {os.path.basename(parc1_path)} to match {os.path.basename(parc2_path)}")
            resampled = ants.resample_image_to_target(
                parc1_ants, parc2_ants, interp_type='nearestNeighbor')
            parc1 = resampled.numpy()
            parc2 = parc2_ants.numpy()
    else:
        # Images are already in the same space
        parc1 = parc1_img.get_fdata()
        parc2 = parc2_img.get_fdata()
    
    # Convert to integer type to ensure label values are preserved
    parc1 = parc1.astype(int)
    parc2 = parc2.astype(int)
    
    # Continue with the existing code
    labels = sorted(set(np.unique(parc1)) | set(np.unique(parc2)))
    labels = [label for label in labels if label != 0]  # Exclude background

    print(f"\nDice scores per label:\n{'Label':<8}{'Region':<40}{'Dice Score':<10}")
    print("-" * 65)

    with open(output_csv_path, mode='w', newline='') as csv_file:
        writer = csv.writer(csv_file)
        writer.writerow(["Label", "Region", "Dice Score"])

        for label in labels:
            mask1 = (parc1 == label)
            mask2 = (parc2 == label)
            dice = dice_score(mask1, mask2)
            region = FREESURFER_LABELS.get(label, "Unknown Region")
            print(f"{label:<8}{region:<40}{dice:.4f}")
            writer.writerow([label, region, f"{dice:.4f}"])

    print(f"\nDice scores with region names saved to: {output_csv_path}")

def print_help():
    """Print help message for dice-compare command."""
    # ANSI color codes
    CYAN = Fore.CYAN
    GREEN = Fore.GREEN
    YELLOW = Fore.YELLOW
    BLUE = Fore.BLUE
    MAGENTA = Fore.MAGENTA
    BOLD = Style.BRIGHT
    RESET = Style.RESET_ALL

    help_text = f"""
    {CYAN}{BOLD}╔════════════════════════════════════════════════════════════════╗
    ║                        DICE COMPARISON                         ║
    ╚════════════════════════════════════════════════════════════════╝{RESET}

    This tool compares two brain parcellation images and calculates the Dice 
    similarity coefficient for each anatomical label. It provides quantitative 
    assessment of registration or segmentation accuracy.

    {CYAN}{BOLD}────────────────────────── USAGE ──────────────────────────{RESET}
      lamar dice-compare {GREEN}[options]{RESET}

    {CYAN}{BOLD}─────────────────── REQUIRED ARGUMENTS ───────────────────{RESET}
      {YELLOW}--ref{RESET} PATH  : Reference parcellation image (.nii.gz)
      {YELLOW}--reg{RESET} PATH  : Registered parcellation image to compare (.nii.gz)
      {YELLOW}--out{RESET} PATH  : Output CSV file for Dice scores (.csv)

    {CYAN}{BOLD}────────────────── EXAMPLE USAGE ────────────────────────{RESET}

    {BLUE}# Calculate Dice scores between reference and registered parcellations{RESET}
    lamar dice-compare {YELLOW}--ref{RESET} fixed_parc.nii.gz {YELLOW}--reg{RESET} registered_parc.nii.gz \\
      {YELLOW}--out{RESET} dice_scores.csv

    {CYAN}{BOLD}────────────────────────── NOTES ───────────────────────{RESET}
    {MAGENTA}•{RESET} Higher Dice scores indicate better spatial overlap between regions
    {MAGENTA}•{RESET} Scores range from 0 (no overlap) to 1 (perfect overlap)
    {MAGENTA}•{RESET} Results include label numbers and anatomical region names
    {MAGENTA}•{RESET} Both FreeSurfer subcortical and Desikan-Killiany cortical regions are supported
    {MAGENTA}•{RESET} Evaluation is performed for each brain region individually
    """
    print(help_text)

def main():
    """Entry point for command-line use"""
    # Check if no arguments were provided or help requested
    if len(sys.argv) == 1 or "-h" in sys.argv or "--help" in sys.argv:
        print_help()
        sys.exit(0)
        
    parser = argparse.ArgumentParser(description="Compute Dice score between two parcellation images.")
    parser.add_argument("--ref", required=True, help="Path to reference parcellation image")
    parser.add_argument("--reg", required=True, help="Path to registered parcellation image")
    parser.add_argument("--out", required=True, help="Output CSV file path")
    args = parser.parse_args()

    compare_parcellations_dice(args.ref, args.reg, args.out)

if __name__ == "__main__":
    main()



"""Utility for splitting large files into chunks and reassembling them."""

import os
import shutil


def split_file(input_path, chunk_size_mb=95, output_dir=None):
    """Split a large file into chunks under the specified size.

    Args:
        input_path: Path to the file to split
        chunk_size_mb: Maximum size of each chunk in MB
        output_dir: Directory to store chunks (default: same as input file)

    Returns:
        List of chunk paths
    """
    if output_dir is None:
        output_dir = os.path.dirname(input_path)

    chunk_size = chunk_size_mb * 1024 * 1024  # Convert MB to bytes
    base_name = os.path.basename(input_path)
    chunk_paths = []

    # Create output directory if it doesn't exist
    os.makedirs(output_dir, exist_ok=True)

    with open(input_path, "rb") as f_in:
        chunk_num = 0
        while True:
            chunk_data = f_in.read(chunk_size)
            if not chunk_data:
                break

            chunk_path = os.path.join(output_dir, f"{base_name}.{chunk_num:03d}")
            chunk_paths.append(chunk_path)

            with open(chunk_path, "wb") as f_out:
                f_out.write(chunk_data)

            print(f"Created chunk {chunk_num}: {chunk_path}")
            chunk_num += 1

    return chunk_paths


def reassemble_file(output_path, pattern=None, chunks_dir=None):
    """Reassemble file from chunks.

    Args:
        output_path: Path where the reassembled file will be saved
        pattern: Base pattern to look for chunks. If None, uses output_path basename.
        chunks_dir: Directory containing chunks. If None, uses output_path directory.

    Returns:
        True if reassembly was successful, False otherwise
    """
    if chunks_dir is None:
        chunks_dir = os.path.dirname(output_path)

    if pattern is None:
        pattern = os.path.basename(output_path)

    # Get all chunks sorted by number
    chunks = [f for f in os.listdir(chunks_dir) if f.startswith(pattern) and "." in f]
    chunks.sort(key=lambda x: int(x.split(".")[-1]))

    if not chunks:
        print(f"No chunks found matching pattern {pattern} in {chunks_dir}")
        return False

    # Create directory if it doesn't exist
    os.makedirs(os.path.dirname(output_path), exist_ok=True)

    # Reassemble the file
    with open(output_path, "wb") as f_out:
        for chunk in chunks:
            chunk_path = os.path.join(chunks_dir, chunk)
            with open(chunk_path, "rb") as f_in:
                shutil.copyfileobj(f_in, f_out)

    print(f"Successfully reassembled {output_path} from {len(chunks)} chunks")
    return True


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Split or reassemble large files")
    subparsers = parser.add_subparsers(dest="command", help="Command to execute")

    # Split command
    split_parser = subparsers.add_parser("split", help="Split a file into chunks")
    split_parser.add_argument("input_path", help="Path to the file to split")
    split_parser.add_argument(
        "--size", type=int, default=95, help="Maximum size of each chunk in MB"
    )
    split_parser.add_argument("--output-dir", help="Directory to store chunks")

    # Reassemble command
    reassemble_parser = subparsers.add_parser(
        "reassemble", help="Reassemble chunks into a file"
    )
    reassemble_parser.add_argument(
        "output_path", help="Path where the reassembled file will be saved"
    )
    reassemble_parser.add_argument("--pattern", help="Base pattern to look for chunks")
    reassemble_parser.add_argument("--chunks-dir", help="Directory containing chunks")

    args = parser.parse_args()

    if args.command == "split":
        split_file(args.input_path, args.size, args.output_dir)
    elif args.command == "reassemble":
        reassemble_file(args.output_path, args.pattern, args.chunks_dir)
    else:
        parser.print_help()



#!/usr/bin/env python3
"""
Full pipeline test for LaMAR.

This script runs a complete registration workflow using the example data
and verifies that all output files are created correctly.
"""

import os
import sys
import subprocess
import nibabel as nib
import shutil
import time
import argparse
import numpy as np
from scipy.ndimage import zoom


def downsample_image(input_path, output_path, factor=4):
    """Downsample a NIfTI image by the specified factor."""
    print(f"Downsampling {os.path.basename(input_path)} by factor {factor}...")

    img = nib.load(input_path)
    data = img.get_fdata()
    affine = img.affine.copy()

    # Calculate zoom factors for each dimension (inverse of downsample factor)
    zoom_factor = [1 / factor] * 3

    # Downsample image
    downsampled_data = zoom(data, zoom_factor, order=1)

    # Adjust affine matrix for the new voxel size
    scaling = np.diag([factor, factor, factor, 1.0])
    new_affine = affine @ scaling

    # Create and save the downsampled image
    new_img = nib.Nifti1Image(downsampled_data, new_affine, img.header)
    nib.save(new_img, output_path)
    print(f"Saved downsampled image to {output_path}")
    return output_path


def test_full_pipeline(quick=False):
    """Run LaMAR pipeline on example data and check outputs."""
    print("\n=== LaMAR Full Pipeline Test ===")

    # Define paths
    example_dir = os.path.join(
        os.path.dirname(os.path.dirname(__file__)), "example_data"
    )
    output_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), "test_output")
    temp_dir = os.path.join(output_dir, "temp")

    # Create output directories
    os.makedirs(output_dir, exist_ok=True)
    os.makedirs(temp_dir, exist_ok=True)

    # Original image paths
    orig_moving_file = os.path.join(
        example_dir, "sub-HC001_ses-02_space-dwi_desc-b0.nii.gz"
    )
    orig_fixed_file = os.path.join(example_dir, "sub-HC001_ses-01_T1w.nii.gz")

    # Check if example data exists
    if not os.path.exists(orig_moving_file):
        print(f"ERROR: Moving image not found: {orig_moving_file}")
        return False
    if not os.path.exists(orig_fixed_file):
        print(f"ERROR: Fixed image not found: {orig_fixed_file}")
        return False

    # Downsample images
    moving_file = os.path.join(temp_dir, "moving_downsampled.nii.gz")
    fixed_file = os.path.join(temp_dir, "fixed_downsampled.nii.gz")

    downsample_image(orig_moving_file, moving_file)
    downsample_image(orig_fixed_file, fixed_file)

    # Define output files
    output_image = os.path.join(output_dir, "sub-001_dwi_in_T1w.nii.gz")
    moving_parc = os.path.join(output_dir, "sub-001_dwi_parc.nii.gz")
    fixed_parc = os.path.join(output_dir, "sub-001_T1w_parc.nii.gz")
    registered_parc = os.path.join(output_dir, "sub-001_dwi_reg_parc.nii.gz")
    affine_file = os.path.join(output_dir, "dwi_to_T1w_affine.mat")
    warp_file = os.path.join(output_dir, "dwi_to_T1w_warp.nii.gz")
    inverse_warp = os.path.join(output_dir, "T1w_to_dwi_warp.nii.gz")
    qc_csv = os.path.join(output_dir, "dice_scores.csv")

    # Clean previous outputs
    for f in [
        output_image,
        moving_parc,
        fixed_parc,
        registered_parc,
        affine_file,
        warp_file,
        inverse_warp,
        qc_csv,
    ]:
        if os.path.exists(f):
            if os.path.isdir(f):
                shutil.rmtree(f)
            else:
                os.remove(f)

    # Build the LaMAR command
    cmd = [
        "lamareg",
        "register",
        "--moving",
        moving_file,
        "--fixed",
        fixed_file,
        "--output",
        output_image,
        "--moving-parc",
        moving_parc,
        "--fixed-parc",
        fixed_parc,
        "--registered-parc",
        registered_parc,
        "--affine",
        affine_file,
        "--warpfield",
        warp_file,
        "--inverse-warpfield",
        inverse_warp,
        "--qc-csv",
        qc_csv,
        "--synthseg-threads",
        "1",  # Lower thread count for CI
        "--ants-threads",
        "1",  # Lower thread count for CI
    ]

    print("\nRunning LaMAR pipeline...")
    print(" ".join(cmd))
    start_time = time.time()

    try:
        # Run the pipeline
        result = subprocess.run(cmd)
        if result.returncode != 0:
            print(f"ERROR: Command failed with return code {result.returncode}")
            print("STDOUT:", result.stdout)
            print("STDERR:", result.stderr)
            return False

        print(result.stdout)
        if result.stderr:
            print("STDERR:", result.stderr)

        # Check if expected output files exist
        expected_files = [
            output_image,
            moving_parc,
            fixed_parc,
            registered_parc,
            affine_file,
            warp_file,
            inverse_warp,
        ]

        missing_files = []
        for f in expected_files:
            if not os.path.exists(f):
                missing_files.append(f)

        if missing_files:
            print("ERROR: The following output files were not created:")
            for f in missing_files:
                print(f"  - {f}")
            return False

        # Verify outputs
        try:
            # Check that images can be loaded
            nib.load(output_image)
            nib.load(moving_parc)
            nib.load(fixed_parc)
            nib.load(registered_parc)

            # Check that transform files are not empty
            if os.path.getsize(affine_file) == 0:
                print(f"ERROR: Affine file is empty: {affine_file}")
                return False
            if os.path.getsize(warp_file) == 0:
                print(f"ERROR: Warp file is empty: {warp_file}")
                return False

            # Check Dice scores if available
            if os.path.exists(qc_csv):
                print("\nRegistration quality metrics:")
                with open(qc_csv, "r") as f:
                    print(f.read())

        except Exception as e:
            print(f"ERROR: Failed to verify output files: {e}")
            return False

        elapsed_time = time.time() - start_time
        print(f"\nTest completed successfully in {elapsed_time:.1f} seconds!")
        return True

    except Exception as e:
        print(f"ERROR: Test failed with exception: {e}")
        return False


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run a full pipeline test for LaMAR")
    args = parser.parse_args()

    success = test_full_pipeline()
    sys.exit(0 if success else 1)



"""
tensorflow/keras utilities for the neuron project

If you use this code, please cite
Dalca AV, Guttag J, Sabuncu MR
Anatomical Priors in Convolutional Networks for Unsupervised Biomedical Segmentation,
CVPR 2018

Contact: adalca [at] csail [dot] mit [dot] edu
License: GPLv3
"""

import sys

from lamareg.ext.neuron import layers

# third party
import numpy as np
import tensorflow as tf
import keras
import keras.layers as KL
from keras.models import Model
import keras.backend as K


def unet(
    nb_features,
    input_shape,
    nb_levels,
    conv_size,
    nb_labels,
    name="unet",
    prefix=None,
    feat_mult=1,
    pool_size=2,
    use_logp=True,
    padding="same",
    dilation_rate_mult=1,
    activation="elu",
    skip_n_concatenations=0,
    use_residuals=False,
    final_pred_activation="softmax",
    nb_conv_per_level=1,
    add_prior_layer=False,
    layer_nb_feats=None,
    conv_dropout=0,
    batch_norm=None,
    input_model=None,
):
    """
    unet-style keras model with an overdose of parametrization.

    Parameters:
        nb_features: the number of features at each convolutional level
            see below for `feat_mult` and `layer_nb_feats` for modifiers to this number
        input_shape: input layer shape, vector of size ndims + 1 (nb_channels)
        conv_size: the convolution kernel size
        nb_levels: the number of Unet levels (number of downsamples) in the "encoder"
            (e.g. 4 would give you 4 levels in encoder, 4 in decoder)
        nb_labels: number of output channels
        name (default: 'unet'): the name of the network
        prefix (default: `name` value): prefix to be added to layer names
        feat_mult (default: 1) multiple for `nb_features` as we go down the encoder levels.
            e.g. feat_mult of 2 and nb_features of 16 would yield 32 features in the
            second layer, 64 features in the third layer, etc.
        pool_size (default: 2): max pooling size (integer or list if specifying per dimension)
        skip_n_concatenations=0: enabled to skip concatenation links between contracting and expanding paths for the n
            top levels.
        use_logp:
        padding:
        dilation_rate_mult:
        activation:
        use_residuals:
        final_pred_activation:
        nb_conv_per_level:
        add_prior_layer:
        skip_n_concatenations:
        layer_nb_feats: list of the number of features for each layer. Automatically used if specified
        conv_dropout: dropout probability
        batch_norm:
        input_model: concatenate the provided input_model to this current model.
            Only the first output of input_model is used.
    """

    # naming
    model_name = name
    if prefix is None:
        prefix = model_name

    # volume size data
    ndims = len(input_shape) - 1
    if isinstance(pool_size, int):
        pool_size = (pool_size,) * ndims

    # get encoding model
    enc_model = conv_enc(
        nb_features,
        input_shape,
        nb_levels,
        conv_size,
        name=model_name,
        prefix=prefix,
        feat_mult=feat_mult,
        pool_size=pool_size,
        padding=padding,
        dilation_rate_mult=dilation_rate_mult,
        activation=activation,
        use_residuals=use_residuals,
        nb_conv_per_level=nb_conv_per_level,
        layer_nb_feats=layer_nb_feats,
        conv_dropout=conv_dropout,
        batch_norm=batch_norm,
        input_model=input_model,
    )

    # get decoder
    # use_skip_connections=True makes it a u-net
    lnf = (
        layer_nb_feats[(nb_levels * nb_conv_per_level) :]
        if layer_nb_feats is not None
        else None
    )
    dec_model = conv_dec(
        nb_features,
        [],
        nb_levels,
        conv_size,
        nb_labels,
        name=model_name,
        prefix=prefix,
        feat_mult=feat_mult,
        pool_size=pool_size,
        use_skip_connections=True,
        skip_n_concatenations=skip_n_concatenations,
        padding=padding,
        dilation_rate_mult=dilation_rate_mult,
        activation=activation,
        use_residuals=use_residuals,
        final_pred_activation="linear" if add_prior_layer else final_pred_activation,
        nb_conv_per_level=nb_conv_per_level,
        batch_norm=batch_norm,
        layer_nb_feats=lnf,
        conv_dropout=conv_dropout,
        input_model=enc_model,
    )
    final_model = dec_model

    if add_prior_layer:
        final_model = add_prior(
            dec_model,
            [*input_shape[:-1], nb_labels],
            name=model_name + "_prior",
            use_logp=use_logp,
            final_pred_activation=final_pred_activation,
        )

    return final_model


def ae(
    nb_features,
    input_shape,
    nb_levels,
    conv_size,
    nb_labels,
    enc_size,
    name="ae",
    feat_mult=1,
    pool_size=2,
    padding="same",
    activation="elu",
    use_residuals=False,
    nb_conv_per_level=1,
    batch_norm=None,
    enc_batch_norm=None,
    ae_type="conv",  # 'dense', or 'conv'
    enc_lambda_layers=None,
    add_prior_layer=False,
    use_logp=True,
    conv_dropout=0,
    include_mu_shift_layer=False,
    single_model=False,  # whether to return a single model, or a tuple of models that can be stacked.
    final_pred_activation="softmax",
    do_vae=False,
    input_model=None,
):
    """Convolutional Auto-Encoder. Optionally Variational (if do_vae is set to True)."""

    # naming
    model_name = name

    # volume size data
    ndims = len(input_shape) - 1
    if isinstance(pool_size, int):
        pool_size = (pool_size,) * ndims

    # get encoding model
    enc_model = conv_enc(
        nb_features,
        input_shape,
        nb_levels,
        conv_size,
        name=model_name,
        feat_mult=feat_mult,
        pool_size=pool_size,
        padding=padding,
        activation=activation,
        use_residuals=use_residuals,
        nb_conv_per_level=nb_conv_per_level,
        conv_dropout=conv_dropout,
        batch_norm=batch_norm,
        input_model=input_model,
    )

    # middle AE structure
    if single_model:
        in_input_shape = None
        in_model = enc_model
    else:
        in_input_shape = enc_model.output.shape.as_list()[1:]
        in_model = None
    mid_ae_model = single_ae(
        enc_size,
        in_input_shape,
        conv_size=conv_size,
        name=model_name,
        ae_type=ae_type,
        input_model=in_model,
        batch_norm=enc_batch_norm,
        enc_lambda_layers=enc_lambda_layers,
        include_mu_shift_layer=include_mu_shift_layer,
        do_vae=do_vae,
    )

    # decoder
    if single_model:
        in_input_shape = None
        in_model = mid_ae_model
    else:
        in_input_shape = mid_ae_model.output.shape.as_list()[1:]
        in_model = None
    dec_model = conv_dec(
        nb_features,
        in_input_shape,
        nb_levels,
        conv_size,
        nb_labels,
        name=model_name,
        feat_mult=feat_mult,
        pool_size=pool_size,
        use_skip_connections=False,
        padding=padding,
        activation=activation,
        use_residuals=use_residuals,
        final_pred_activation="linear",
        nb_conv_per_level=nb_conv_per_level,
        batch_norm=batch_norm,
        conv_dropout=conv_dropout,
        input_model=in_model,
    )

    if add_prior_layer:
        dec_model = add_prior(
            dec_model,
            [*input_shape[:-1], nb_labels],
            name=model_name,
            prefix=model_name + "_prior",
            use_logp=use_logp,
            final_pred_activation=final_pred_activation,
        )

    if single_model:
        return dec_model
    else:
        return dec_model, mid_ae_model, enc_model


def conv_enc(
    nb_features,
    input_shape,
    nb_levels,
    conv_size,
    name=None,
    prefix=None,
    feat_mult=1,
    pool_size=2,
    dilation_rate_mult=1,
    padding="same",
    activation="elu",
    layer_nb_feats=None,
    use_residuals=False,
    nb_conv_per_level=2,
    conv_dropout=0,
    batch_norm=None,
    input_model=None,
):
    """Fully Convolutional Encoder"""

    # naming
    model_name = name
    if prefix is None:
        prefix = model_name

    # first layer: input
    name = "%s_input" % prefix
    if input_model is None:
        input_tensor = KL.Input(shape=input_shape, name=name)
        last_tensor = input_tensor
    else:
        input_tensor = input_model.inputs
        last_tensor = input_model.outputs
        if isinstance(last_tensor, list):
            last_tensor = last_tensor[0]

    # volume size data
    ndims = len(input_shape) - 1
    if isinstance(pool_size, int):
        pool_size = (pool_size,) * ndims

    # prepare layers
    convL = getattr(KL, "Conv%dD" % ndims)
    conv_kwargs = {
        "padding": padding,
        "activation": activation,
        "data_format": "channels_last",
    }
    maxpool = getattr(KL, "MaxPooling%dD" % ndims)

    # down arm:
    # add nb_levels of conv + ReLu + conv + ReLu. Pool after each of first nb_levels - 1 layers
    lfidx = 0  # level feature index
    for level in range(nb_levels):
        lvl_first_tensor = last_tensor
        nb_lvl_feats = np.round(nb_features * feat_mult**level).astype(int)
        conv_kwargs["dilation_rate"] = dilation_rate_mult**level

        for conv in range(
            nb_conv_per_level
        ):  # does several conv per level, max pooling applied at the end
            if layer_nb_feats is not None:  # None or List of all the feature numbers
                nb_lvl_feats = layer_nb_feats[lfidx]
                lfidx += 1

            name = "%s_conv_downarm_%d_%d" % (prefix, level, conv)
            if conv < (nb_conv_per_level - 1) or (not use_residuals):
                last_tensor = convL(nb_lvl_feats, conv_size, **conv_kwargs, name=name)(
                    last_tensor
                )
            else:  # no activation
                last_tensor = convL(
                    nb_lvl_feats, conv_size, padding=padding, name=name
                )(last_tensor)

            if conv_dropout > 0:
                # conv dropout along feature space only
                name = "%s_dropout_downarm_%d_%d" % (prefix, level, conv)
                noise_shape = [None, *[1] * ndims, nb_lvl_feats]
                last_tensor = KL.Dropout(
                    conv_dropout, noise_shape=noise_shape, name=name
                )(last_tensor)

        if use_residuals:
            convarm_layer = last_tensor

            # the "add" layer is the original input
            # However, it may not have the right number of features to be added
            nb_feats_in = lvl_first_tensor.get_shape()[-1]
            nb_feats_out = convarm_layer.get_shape()[-1]
            add_layer = lvl_first_tensor
            if nb_feats_in > 1 and nb_feats_out > 1 and (nb_feats_in != nb_feats_out):
                name = "%s_expand_down_merge_%d" % (prefix, level)
                last_tensor = convL(nb_lvl_feats, conv_size, **conv_kwargs, name=name)(
                    lvl_first_tensor
                )
                add_layer = last_tensor

                if conv_dropout > 0:
                    noise_shape = [None, *[1] * ndims, nb_lvl_feats]
                    convarm_layer = KL.Dropout(conv_dropout, noise_shape=noise_shape)(
                        last_tensor
                    )

            name = "%s_res_down_merge_%d" % (prefix, level)
            last_tensor = KL.add([add_layer, convarm_layer], name=name)

            name = "%s_res_down_merge_act_%d" % (prefix, level)
            last_tensor = KL.Activation(activation, name=name)(last_tensor)

        if batch_norm is not None:
            name = "%s_bn_down_%d" % (prefix, level)
            last_tensor = KL.BatchNormalization(axis=batch_norm, name=name)(last_tensor)

        # max pool if we're not at the last level
        if level < (nb_levels - 1):
            name = "%s_maxpool_%d" % (prefix, level)
            last_tensor = maxpool(pool_size=pool_size, name=name, padding=padding)(
                last_tensor
            )

    # create the model and return
    model = Model(inputs=input_tensor, outputs=[last_tensor], name=model_name)
    return model


def conv_dec(
    nb_features,
    input_shape,
    nb_levels,
    conv_size,
    nb_labels,
    name=None,
    prefix=None,
    feat_mult=1,
    pool_size=2,
    use_skip_connections=False,
    skip_n_concatenations=0,
    padding="same",
    dilation_rate_mult=1,
    activation="elu",
    use_residuals=False,
    final_pred_activation="softmax",
    nb_conv_per_level=2,
    layer_nb_feats=None,
    batch_norm=None,
    conv_dropout=0,
    input_model=None,
):
    """Fully Convolutional Decoder"""

    # naming
    model_name = name
    if prefix is None:
        prefix = model_name

    # if using skip connections, make sure need to use them.
    if use_skip_connections:
        assert (
            input_model is not None
        ), "is using skip connections, tensors dictionary is required"

    # first layer: input
    input_name = "%s_input" % prefix
    if input_model is None:
        input_tensor = KL.Input(shape=input_shape, name=input_name)
        last_tensor = input_tensor
    else:
        input_tensor = input_model.input
        last_tensor = input_model.output
        if isinstance(last_tensor, list):
            last_tensor = last_tensor[0]
            input_shape = list(last_tensor.shape[1:])
        else:
            input_shape = last_tensor.shape.as_list()[1:]

    # vol size info
    ndims = len(input_shape) - 1
    if isinstance(pool_size, int):
        if ndims > 1:
            pool_size = (pool_size,) * ndims

    # prepare layers
    convL = getattr(KL, "Conv%dD" % ndims)
    conv_kwargs = {"padding": padding, "activation": activation}
    upsample = getattr(KL, "UpSampling%dD" % ndims)

    # up arm:
    # nb_levels - 1 layers of Deconvolution3D
    #    (approx via up + conv + ReLu) + merge + conv + ReLu + conv + ReLu
    lfidx = 0
    for level in range(nb_levels - 1):
        nb_lvl_feats = np.round(
            nb_features * feat_mult ** (nb_levels - 2 - level)
        ).astype(int)
        conv_kwargs["dilation_rate"] = dilation_rate_mult ** (nb_levels - 2 - level)

        # upsample matching the max pooling layers size
        name = "%s_up_%d" % (prefix, nb_levels + level)
        last_tensor = upsample(size=pool_size, name=name)(last_tensor)
        up_tensor = last_tensor

        # merge layers combining previous layer
        if use_skip_connections & (level < (nb_levels - skip_n_concatenations - 1)):
            conv_name = "%s_conv_downarm_%d_%d" % (
                prefix,
                nb_levels - 2 - level,
                nb_conv_per_level - 1,
            )
            cat_tensor = input_model.get_layer(conv_name).output
            name = "%s_merge_%d" % (prefix, nb_levels + level)
            last_tensor = KL.concatenate(
                [cat_tensor, last_tensor], axis=ndims + 1, name=name
            )

        # convolution layers
        for conv in range(nb_conv_per_level):
            if layer_nb_feats is not None:
                nb_lvl_feats = layer_nb_feats[lfidx]
                lfidx += 1

            name = "%s_conv_uparm_%d_%d" % (prefix, nb_levels + level, conv)
            if conv < (nb_conv_per_level - 1) or (not use_residuals):
                last_tensor = convL(nb_lvl_feats, conv_size, **conv_kwargs, name=name)(
                    last_tensor
                )
            else:
                last_tensor = convL(
                    nb_lvl_feats, conv_size, padding=padding, name=name
                )(last_tensor)

            if conv_dropout > 0:
                name = "%s_dropout_uparm_%d_%d" % (prefix, level, conv)
                noise_shape = [None, *[1] * ndims, nb_lvl_feats]
                last_tensor = KL.Dropout(
                    conv_dropout, noise_shape=noise_shape, name=name
                )(last_tensor)

        # residual block
        if use_residuals:

            # the "add" layer is the original input
            # However, it may not have the right number of features to be added
            add_layer = up_tensor
            nb_feats_in = add_layer.get_shape()[-1]
            nb_feats_out = last_tensor.get_shape()[-1]
            if nb_feats_in > 1 and nb_feats_out > 1 and (nb_feats_in != nb_feats_out):
                name = "%s_expand_up_merge_%d" % (prefix, level)
                add_layer = convL(nb_lvl_feats, conv_size, **conv_kwargs, name=name)(
                    add_layer
                )

                if conv_dropout > 0:
                    noise_shape = [None, *[1] * ndims, nb_lvl_feats]
                    last_tensor = KL.Dropout(conv_dropout, noise_shape=noise_shape)(
                        last_tensor
                    )

            name = "%s_res_up_merge_%d" % (prefix, level)
            last_tensor = KL.add([last_tensor, add_layer], name=name)

            name = "%s_res_up_merge_act_%d" % (prefix, level)
            last_tensor = KL.Activation(activation, name=name)(last_tensor)

        if batch_norm is not None:
            name = "%s_bn_up_%d" % (prefix, level)
            last_tensor = KL.BatchNormalization(axis=batch_norm, name=name)(last_tensor)

    # Compute likelihood prediction (no activation yet)
    name = "%s_likelihood" % prefix
    last_tensor = convL(nb_labels, 1, activation=None, name=name)(last_tensor)
    like_tensor = last_tensor

    # output prediction layer
    # we use a softmax to compute P(L_x|I) where x is each location
    if final_pred_activation == "softmax":
        name = "%s_prediction" % prefix
        softmax_lambda_fcn = lambda x: keras.activations.softmax(x, axis=ndims + 1)
        pred_tensor = KL.Lambda(softmax_lambda_fcn, name=name)(last_tensor)

    # otherwise create a layer that does nothing.
    else:
        name = "%s_prediction" % prefix
        pred_tensor = KL.Activation("linear", name=name)(like_tensor)

    # create the model and return
    model = Model(inputs=input_tensor, outputs=pred_tensor, name=model_name)
    return model


def add_prior(
    input_model,
    prior_shape,
    name="prior_model",
    prefix=None,
    use_logp=True,
    final_pred_activation="softmax",
):
    """
    Append post-prior layer to a given model
    """

    # naming
    model_name = name
    if prefix is None:
        prefix = model_name

    # prior input layer
    prior_input_name = "%s-input" % prefix
    prior_tensor = KL.Input(shape=prior_shape, name=prior_input_name)
    prior_tensor_input = prior_tensor
    like_tensor = input_model.output

    # operation varies depending on whether we log() prior or not.
    if use_logp:
        print(
            "Breaking change: use_logp option now requires log input!", file=sys.stderr
        )
        merge_op = KL.add

    else:
        # using sigmoid to get the likelihood values between 0 and 1
        # note: they won't add up to 1.
        name = "%s_likelihood_sigmoid" % prefix
        like_tensor = KL.Activation("sigmoid", name=name)(like_tensor)
        merge_op = KL.multiply

    # merge the likelihood and prior layers into posterior layer
    name = "%s_posterior" % prefix
    post_tensor = merge_op([prior_tensor, like_tensor], name=name)

    # output prediction layer
    # we use a softmax to compute P(L_x|I) where x is each location
    pred_name = "%s_prediction" % prefix
    if final_pred_activation == "softmax":
        assert use_logp, "cannot do softmax when adding prior via P()"
        print(
            "using final_pred_activation %s for %s"
            % (final_pred_activation, model_name)
        )
        softmax_lambda_fcn = lambda x: keras.activations.softmax(x, axis=-1)
        pred_tensor = KL.Lambda(softmax_lambda_fcn, name=pred_name)(post_tensor)

    else:
        pred_tensor = KL.Activation("linear", name=pred_name)(post_tensor)

    # create the model
    model_inputs = [*input_model.inputs, prior_tensor_input]
    model = Model(inputs=model_inputs, outputs=[pred_tensor], name=model_name)

    # compile
    return model


def single_ae(
    enc_size,
    input_shape,
    name="single_ae",
    prefix=None,
    ae_type="dense",  # 'dense', or 'conv'
    conv_size=None,
    input_model=None,
    enc_lambda_layers=None,
    batch_norm=True,
    padding="same",
    activation=None,
    include_mu_shift_layer=False,
    do_vae=False,
):
    """single-layer Autoencoder (i.e. input - encoding - output"""

    # naming
    model_name = name
    if prefix is None:
        prefix = model_name

    if enc_lambda_layers is None:
        enc_lambda_layers = []

    # prepare input
    input_name = "%s_input" % prefix
    if input_model is None:
        assert input_shape is not None, "input_shape of input_model is necessary"
        input_tensor = KL.Input(shape=input_shape, name=input_name)
        last_tensor = input_tensor
    else:
        input_tensor = input_model.input
        last_tensor = input_model.output
        input_shape = last_tensor.shape.as_list()[1:]
    input_nb_feats = last_tensor.shape.as_list()[-1]

    # prepare conv type based on input
    ndims = len(input_shape) - 1
    if ae_type == "conv":
        convL = getattr(KL, "Conv%dD" % ndims)
        assert conv_size is not None, "with conv ae, need conv_size"
        conv_kwargs = {"padding": padding, "activation": activation}
        enc_size_str = None

    # if want to go through a dense layer in the middle of the U, need to:
    # - flatten last layer if not flat
    # - do dense encoding and decoding
    # - unflatten (reshape spatially) at end
    else:  # ae_type == 'dense'
        if len(input_shape) > 1:
            name = "%s_ae_%s_down_flat" % (prefix, ae_type)
            last_tensor = KL.Flatten(name=name)(last_tensor)
        convL = conv_kwargs = None
        assert len(enc_size) == 1, "enc_size should be of length 1 for dense layer"
        enc_size_str = "".join(["%d_" % d for d in enc_size])[:-1]

    # recall this layer
    pre_enc_layer = last_tensor

    # encoding layer
    if ae_type == "dense":
        name = "%s_ae_mu_enc_dense_%s" % (prefix, enc_size_str)
        last_tensor = KL.Dense(enc_size[0], name=name)(pre_enc_layer)

    else:  # convolution

        # convolve then resize. enc_size should be [nb_dim1, nb_dim2, ..., nb_feats]
        assert len(enc_size) == len(
            input_shape
        ), "encoding size does not match input shape %d %d" % (
            len(enc_size),
            len(input_shape),
        )

        if (
            list(enc_size)[:-1] != list(input_shape)[:-1]
            and all([f is not None for f in input_shape[:-1]])
            and all([f is not None for f in enc_size[:-1]])
        ):

            name = "%s_ae_mu_enc_conv" % prefix
            last_tensor = convL(enc_size[-1], conv_size, name=name, **conv_kwargs)(
                pre_enc_layer
            )

            name = "%s_ae_mu_enc" % prefix
            zf = [
                enc_size[:-1][f] / last_tensor.shape.as_list()[1:-1][f]
                for f in range(len(enc_size) - 1)
            ]
            last_tensor = layers.Resize(zoom_factor=zf, name=name)(last_tensor)

        elif enc_size[-1] is None:  # convolutional, but won't tell us bottleneck
            name = "%s_ae_mu_enc" % prefix
            last_tensor = KL.Lambda(lambda x: x, name=name)(pre_enc_layer)

        else:
            name = "%s_ae_mu_enc" % prefix
            last_tensor = convL(enc_size[-1], conv_size, name=name, **conv_kwargs)(
                pre_enc_layer
            )

    if include_mu_shift_layer:
        # shift
        name = "%s_ae_mu_shift" % prefix
        last_tensor = layers.LocalBias(name=name)(last_tensor)

    # encoding clean-up layers
    for layer_fcn in enc_lambda_layers:
        lambda_name = layer_fcn.__name__
        name = "%s_ae_mu_%s" % (prefix, lambda_name)
        last_tensor = KL.Lambda(layer_fcn, name=name)(last_tensor)

    if batch_norm is not None:
        name = "%s_ae_mu_bn" % prefix
        last_tensor = KL.BatchNormalization(axis=batch_norm, name=name)(last_tensor)

    # have a simple layer that does nothing to have a clear name before sampling
    name = "%s_ae_mu" % prefix
    last_tensor = KL.Lambda(lambda x: x, name=name)(last_tensor)

    # if doing variational AE, will need the sigma layer as well.
    if do_vae:
        mu_tensor = last_tensor

        # encoding layer
        if ae_type == "dense":
            name = "%s_ae_sigma_enc_dense_%s" % (prefix, enc_size_str)
            last_tensor = KL.Dense(enc_size[0], name=name)(pre_enc_layer)

        else:
            if (
                list(enc_size)[:-1] != list(input_shape)[:-1]
                and all([f is not None for f in input_shape[:-1]])
                and all([f is not None for f in enc_size[:-1]])
            ):

                assert (
                    len(enc_size) - 1 == 2
                ), "Sorry, I have not yet implemented non-2D resizing..."
                name = "%s_ae_sigma_enc_conv" % prefix
                last_tensor = convL(enc_size[-1], conv_size, name=name, **conv_kwargs)(
                    pre_enc_layer
                )

                name = "%s_ae_sigma_enc" % prefix
                resize_fn = lambda x: tf.image.resize_bilinear(x, enc_size[:-1])
                last_tensor = KL.Lambda(resize_fn, name=name)(last_tensor)

            elif enc_size[-1] is None:  # convolutional, but won't tell us bottleneck
                name = "%s_ae_sigma_enc" % prefix
                last_tensor = convL(
                    pre_enc_layer.shape.as_list()[-1],
                    conv_size,
                    name=name,
                    **conv_kwargs,
                )(pre_enc_layer)
                # cannot use lambda, then mu and sigma will be same layer.
                # last_tensor = KL.Lambda(lambda x: x, name=name)(pre_enc_layer)

            else:
                name = "%s_ae_sigma_enc" % prefix
                last_tensor = convL(enc_size[-1], conv_size, name=name, **conv_kwargs)(
                    pre_enc_layer
                )

        # encoding clean-up layers
        for layer_fcn in enc_lambda_layers:
            lambda_name = layer_fcn.__name__
            name = "%s_ae_sigma_%s" % (prefix, lambda_name)
            last_tensor = KL.Lambda(layer_fcn, name=name)(last_tensor)

        if batch_norm is not None:
            name = "%s_ae_sigma_bn" % prefix
            last_tensor = KL.BatchNormalization(axis=batch_norm, name=name)(last_tensor)

        # have a simple layer that does nothing to have a clear name before sampling
        name = "%s_ae_sigma" % prefix
        last_tensor = KL.Lambda(lambda x: x, name=name)(last_tensor)

        logvar_tensor = last_tensor

        # VAE sampling
        sampler = _VAESample().sample_z

        name = "%s_ae_sample" % prefix
        last_tensor = KL.Lambda(sampler, name=name)([mu_tensor, logvar_tensor])

    if include_mu_shift_layer:
        # shift
        name = "%s_ae_sample_shift" % prefix
        last_tensor = layers.LocalBias(name=name)(last_tensor)

    # decoding layer
    if ae_type == "dense":
        name = "%s_ae_%s_dec_flat_%s" % (prefix, ae_type, enc_size_str)
        last_tensor = KL.Dense(np.prod(input_shape), name=name)(last_tensor)

        # unflatten if dense method
        if len(input_shape) > 1:
            name = "%s_ae_%s_dec" % (prefix, ae_type)
            last_tensor = KL.Reshape(input_shape, name=name)(last_tensor)

    else:

        if (
            list(enc_size)[:-1] != list(input_shape)[:-1]
            and all([f is not None for f in input_shape[:-1]])
            and all([f is not None for f in enc_size[:-1]])
        ):
            name = "%s_ae_mu_dec" % prefix
            zf = [
                last_tensor.shape.as_list()[1:-1][f] / enc_size[:-1][f]
                for f in range(len(enc_size) - 1)
            ]
            last_tensor = layers.Resize(zoom_factor=zf, name=name)(last_tensor)

        name = "%s_ae_%s_dec" % (prefix, ae_type)
        last_tensor = convL(input_nb_feats, conv_size, name=name, **conv_kwargs)(
            last_tensor
        )

    if batch_norm is not None:
        name = "%s_bn_ae_%s_dec" % (prefix, ae_type)
        last_tensor = KL.BatchNormalization(axis=batch_norm, name=name)(last_tensor)

    # create the model and return
    model = Model(inputs=input_tensor, outputs=[last_tensor], name=model_name)
    return model


###############################################################################
# Helper function
###############################################################################


class _VAESample:
    def __init__(self):
        pass

    def sample_z(self, args):
        mu, log_var = args
        shape = K.shape(mu)
        eps = K.random_normal(shape=shape, mean=0.0, stddev=1.0)
        return mu + K.exp(log_var / 2) * eps



"""
Example script for contrast-agnostic registration using SynthSeg

This script demonstrates a full registration pipeline that uses SynthSeg's brain
parcellation to enable registration between images of different contrasts:

1. Generate parcellations of both input and reference images using SynthSeg
2. Register the parcellations to each other (contrast-agnostic)
3. Apply the resulting transformation to the original input image

This approach is useful for registering images with very different contrasts
(e.g., T1w to T2w, FLAIR to T1w, etc.) where direct intensity-based
registration might fail.
"""

import os
import argparse
import subprocess
import sys
import multiprocessing
import tempfile
DEFAULT_THREADS = multiprocessing.cpu_count()

def lamareg(
    input_image,
    reference_image,
    output_image=None,
    input_parc=None,
    reference_parc=None,
    output_parc=None,
    generate_warpfield=False,
    apply_warpfield=False,
    registration_method="SyNRA",
    affine_file=None,
    warp_file=None,
    inverse_warp_file=None,
    synthseg_threads=DEFAULT_THREADS,
    ants_threads=DEFAULT_THREADS,
    qc_csv=None,
    skip_fixed_parc=False,
    skip_moving_parc=False,
    skip_qc=False,
    disable_robust=False,
    inverse=False,
    secondary_warp_file=None,
    inverse_secondary_warp_file=None
    
):
    """
    Perform contrast-agnostic registration using SynthSeg parcellation.
    """
    # Validate arguments based on the selected workflow
    if generate_warpfield and apply_warpfield:
        raise ValueError(
            "Cannot use both --generate-warpfield and --apply-warpfield at the same time"
        )

    # Validate input files exist
    for input_file in [f for f in [input_image, reference_image] if f is not None]:
        if not os.path.isfile(input_file):
            raise FileNotFoundError(f"Input file not found: {input_file}")

    # Validate thread counts
    if synthseg_threads < 1:
        raise ValueError(
            f"Invalid thread count for SynthSeg: {synthseg_threads}. Must be >= 1"
        )
    if ants_threads < 1:
        raise ValueError(f"Invalid thread count for ANTs: {ants_threads}. Must be >= 1")
    print("skip qc = ", skip_qc)
    # Workflow-specific validation
    if not apply_warpfield:
        # Registration or Generate-warpfield workflow
        if input_image is None:
            raise ValueError("--moving is required for registration")
        if reference_image is None:
            raise ValueError("--fixed is required for registration")
        if input_parc is None:
            raise ValueError("--moving-parc is required for registration")
        if reference_parc is None:
            raise ValueError("--fixed-parc is required for registration")
        if output_parc is None:
            raise ValueError("--registered-parc is required for registration")

        # For normal registration (not generate-warpfield), output image is required
        if not generate_warpfield and output_image is None:
            raise ValueError("--output is required for registration")

        # If generating warpfield, warn if transform files not specified
        if affine_file is None:
            print(
                "Warning: No affine transform file path provided - affine transform will not be saved"
            )
        if warp_file is None:
            print(
                "Warning: No warp field file path provided - warp field will not be saved"
            )
        if inverse_warp_file is None:
            print(
                "Warning: No inverse warp field file path provided - inverse warp field will not be saved"
            )
        if secondary_warp_file is None:
            if not disable_robust:
                print(
                    "Warning: No secondary warp field file path provided - first and second stage warp fields will be composed"
                )
                print(
                    "Warning: Warpfield composition can result in loss of precision and accuracy in the resulting warpfields. Consider providing a secondary warp field file path for improved results if you plan to re-use warpfields."
                )
        if inverse_secondary_warp_file is None:
            if not disable_robust:
                print(
                    "Warning: No inverse secondary warp field file path provided - first and second stage inverse warp fields will be composed"
                )
                print(
                    "Warning: Warpfield composition can result in loss of precision and accuracy in the resulting warpfields. Consider providing a secondary inverse warp field file path for improved results if you plan to re-use warpfields."
                )
    else:
        # Apply-warpfield workflow
        if input_image is None:
            raise ValueError("--moving is required for apply-warpfield")
        if reference_image is None:
            raise ValueError("--fixed is required for apply-warpfield")
        if output_image is None:
            raise ValueError("--output is required for apply-warpfield")
        if affine_file is None:
            raise ValueError("--affine is required for apply-warpfield")
        if warp_file is None:
            raise ValueError("--warpfield is required for apply-warpfield")

        # Validate transform files exist
        for transform_file in [affine_file, warp_file]:
            if not os.path.isfile(transform_file):
                raise FileNotFoundError(f"Transform file not found: {transform_file}")

    # Add QC CSV validation
    if not apply_warpfield and not generate_warpfield:
        # Only relevant for standard registration workflow
        if qc_csv is None:
            # No QC CSV path provided, will use default
            if output_parc is not None:
                default_qc_path = os.path.splitext(output_parc)[0] + "_dice_scores.csv"
                print(
                    f"No QC CSV path provided - Dice scores will be saved to: {default_qc_path}"
                )
        else:
            # QC CSV path provided, check if directory is writable
            qc_dir = os.path.dirname(qc_csv)
            if qc_dir:
                if os.path.exists(qc_dir):
                    if not os.access(qc_dir, os.W_OK):
                        raise PermissionError(
                            f"Cannot write to QC CSV directory: {qc_dir}. Check permissions."
                        )
                else:
                    try:
                        os.makedirs(qc_dir, exist_ok=True)
                    except Exception as e:
                        raise PermissionError(
                            f"Cannot create QC CSV directory: {qc_dir}. Error: {e}"
                        )

    # Create directories for all output files
    for file_path in [
        output_image,
        input_parc,
        reference_parc,
        output_parc,
        affine_file,
        warp_file,
        inverse_warp_file,
        qc_csv,
    ]:
        if file_path is not None:
            output_dir = os.path.dirname(file_path)
            if output_dir:  # Only try to create if there's a directory part
                try:
                    os.makedirs(output_dir, exist_ok=True)
                except PermissionError:
                    raise PermissionError(
                        f"Cannot create output directory: {output_dir}. Check permissions."
                    )

    print(f"Processing input image: {input_image}")
    print(f"Reference image: {reference_image}")
    print(
        f"Using {synthseg_threads} thread(s) for SynthSeg and {ants_threads} thread(s) for ANTs"
    )

    # Create environment with suppressed TensorFlow warnings
    env = os.environ.copy()
    env["TF_CPP_MIN_LOG_LEVEL"] = "3"  # 0=ALL, 1=INFO, 2=WARNING, 3=ERROR
    env["PYTHONWARNINGS"] = "ignore"

    # Set ANTs/ITK thread count
    env["ITK_GLOBAL_DEFAULT_NUMBER_OF_THREADS"] = str(ants_threads)
    env["OMP_NUM_THREADS"] = str(ants_threads)  # OpenMP threads for ANTs

    try:
        # WORKFLOW 1 & 2: Full registration or generate warpfield
        if not apply_warpfield:
            # Step 1: Generate parcellations with SynthSeg if needed
            if not skip_moving_parc:
                print("\n--- Step 1.1: Generating parcellation for input image ---")
                subprocess.run(
                    [
                        "lamareg",
                        "synthseg",
                        "--i",
                        input_image,
                        "--o",
                        input_parc,
                        "--parc",
                        "--cpu",
                        "--robust",
                        "--threads",
                        str(synthseg_threads),  # Use SynthSeg threads
                    ],
                    check=True,
                    env=env,
                )
            else:
                print(
                    f"Skipping parcellation generation for input image: {input_image}"
                )
                print(f"Using provided parcellation: {input_parc}")
            # Check if the input parcellation file exists
            if not os.path.isfile(input_parc):
                raise FileNotFoundError(
                    f"Input parcellation file not found: {input_parc}"
                )

            if not skip_fixed_parc:
                print("\n--- Step 1.2: Generating parcellation for reference image ---")
                subprocess.run(
                    [
                        "lamareg",
                        "synthseg",
                        "--i",
                        reference_image,
                        "--o",
                        reference_parc,
                        "--parc",
                        "--cpu",
                        "--robust",
                        "--threads",
                        str(synthseg_threads),  # Use SynthSeg threads
                    ],
                    check=True,
                    env=env,
                )
            else:
                print(
                    f"Skipping parcellation generation for reference image: {reference_image}"
                )
                print(f"Using provided parcellation: {reference_parc}")
            # Check if the reference parcellation file exists
            if not os.path.isfile(reference_parc):
                raise FileNotFoundError(
                    f"Reference parcellation file not found: {reference_parc}"
                )

            # Step 2: Register parcellations using coregister
            print("\n--- Step 2: Coregistering parcellated images ---")
            cmd = [
                "lamareg",
                "coregister",
                "--fixed",
                reference_parc,
                "--moving",
                input_parc,
                "--registration-method",
                registration_method,
                "--fixed-image",
                reference_image,
            ]

            if output_parc is not None:
                if not disable_robust:
                    with tempfile.NamedTemporaryFile(suffix=f"_tmp_output_parc.nii.gz", delete=False) as tmp_parc:
                        temp_parc_file = tmp_parc.name
                    cmd.extend(["--output", temp_parc_file])
                else:
                    cmd.extend(["--output", output_parc])


            # Only include transform file flags if paths were provided
            if affine_file:
                cmd.extend(["--affine-file", affine_file])

            if warp_file:
                if not disable_robust:
                    if not secondary_warp_file:
                        with tempfile.NamedTemporaryFile(suffix=f"_tmp_warp.nii.gz", delete=False) as tmp_warp:
                            temp_warp_file = tmp_warp.name
                        cmd.extend(["--warp-file", temp_warp_file])
                    else:
                        cmd.extend(["--warp-file", warp_file])
                else:
                    cmd.extend(["--warp-file", warp_file])

            if inverse_warp_file:
                if not disable_robust:
                    if not inverse_secondary_warp_file:
                        with tempfile.NamedTemporaryFile(suffix=f"_tmp_inverse_warp.nii.gz", delete=False) as tmp_inverse_warp:
                            temp_inverse_warp_file = tmp_inverse_warp.name
                        cmd.extend(["--inverse-warp-file", temp_inverse_warp_file])
                    else:
                        cmd.extend(["--inverse-warp-file", inverse_warp_file])
                else:
                    cmd.extend(["--inverse-warp-file", inverse_warp_file])

            subprocess.run(cmd, check=True, env=env)
            if not disable_robust:
                print(
                    "\n--- Step 2.1: Running robust registration for improved accuracy ---"
                )
                robust_cmd = [
                    "lamareg",
                    "coregister",
                    "--fixed",
                    reference_image,
                    "--moving",
                    input_image,
                    "--interpolator",
                    "linear",
                    "--registration-method",
                    "SyNOnly",
                    "--initial-affine-file",
                    affine_file,
                    "--initial-warp-file",
                    temp_warp_file if not secondary_warp_file else warp_file,
                    "--initial-inverse-warp-file",
                    temp_inverse_warp_file if not inverse_secondary_warp_file else inverse_warp_file,
                    "--reg-iterations",
                    "10, 20",
                ]

                if output_image is not None:
                    robust_cmd.extend(["--output", output_image])

                # Only include transform file flags if paths were provided
                if affine_file or not skip_qc:
                    robust_cmd.extend(["--affine-file", affine_file])

                if warp_file or not skip_qc:
                    if secondary_warp_file:
                        robust_cmd.extend(["--warp-file", secondary_warp_file])
                        robust_cmd.extend(["--disable-warp-composition"])
                    else:
                        robust_cmd.extend(["--warp-file", warp_file])

                if inverse_warp_file:
                    if inverse_secondary_warp_file:
                        robust_cmd.extend(["--inverse-warp-file", inverse_secondary_warp_file])
                        robust_cmd.extend(["--disable-inverse-warp-composition"])
                    else:
                        robust_cmd.extend(["--inverse-warp-file", inverse_warp_file])

                subprocess.run(robust_cmd, check=True, env=env)
                try:
                    os.remove(temp_warp_file)
                    os.remove(temp_inverse_warp_file)
                except:
                    pass

                apply_cmd = [
                    "lamareg",
                    "apply-warp",  # Use hyphen instead of underscore
                    "--moving",
                    input_parc,
                    "--fixed",  # Changed from --reference to --fixed
                    reference_image,
                    "--output",
                    output_parc, 
                    "--interpolation",
                    "nearestNeighbor",
                ]

                # Only include transform file flags if files were provided
                apply_cmd.extend(["--affine", affine_file])
                    
                apply_cmd.extend(["--warp", warp_file])

                if secondary_warp_file:
                    apply_cmd.extend(["--secondary-warp", secondary_warp_file])
                
                subprocess.run(apply_cmd, check=True, env=env)
                    
            # Run Dice evaluation after coregistration
            if not skip_qc:


                # If qc_csv is not provided, generate a default path based on output_parc
                dice_output = (
                    qc_csv
                    if qc_csv
                    else output_parc.split('.')[0] + "_dice_scores.csv"
                )

                print(
                    "\n--- Step 3: Calculating Dice scores to evaluate registration quality ---"
                )
                try:
                    from lamareg.scripts.dice_compare import compare_parcellations_dice

                    compare_parcellations_dice(reference_parc, output_parc, dice_output)
                    print(f"Quality control metrics saved to: {dice_output}")
                except FileNotFoundError as e:
                    print(
                        f"Warning: Could not calculate Dice scores - file not found: {e}",
                        file=sys.stderr,
                    )
                except PermissionError as e:
                    print(
                        f"Warning: Could not calculate Dice scores - permission error: {e}",
                        file=sys.stderr,
                    )
                except ImportError as e:
                    print(
                        f"Warning: Could not calculate Dice scores - dice_compare module not found",
                        file=sys.stderr,
                    )
                except Exception as e:
                    print(
                        f"Warning: Could not calculate Dice scores: {e}",
                        file=sys.stderr,
                    )

        # WORKFLOW 1 & 3: Apply transformation to the original input image
        if not generate_warpfield and output_image is not None and (disable_robust or apply_warpfield):
            print(
                "\n--- Step 4: Applying transformation to original input image ---"
            )
            apply_cmd = [
                "lamareg",
                "apply-warp",  # Use hyphen instead of underscore
                "--moving",
                input_image,
                "--fixed",  # Changed from --reference to --fixed
                reference_image,
                "--output",
                output_image, 
            ]

            # Only include transform file flags if files were provided
            if affine_file:
                apply_cmd.extend(["--affine", affine_file])

            if warp_file:
                apply_cmd.extend(["--warp", warp_file])
            
            if secondary_warp_file:
                apply_cmd.extend(["--secondary-warp", secondary_warp_file])

            if inverse:
                apply_cmd.extend(["--inverse"])
            
            subprocess.run(apply_cmd, check=True, env=env)

            print(f"\nSuccess! Registered image saved to: {output_image}")
        elif generate_warpfield:
            success_msg = "\nSuccess! "
            if warp_file:
                success_msg += f"Warp field generated at: {warp_file}"
            if affine_file:
                success_msg += f"\nAffine transformation saved at: {affine_file}"
            print(success_msg)

    except subprocess.CalledProcessError as e:
        print(f"Error during processing: {e}", file=sys.stderr)
        sys.exit(1)


def main():
    """Entry point for command-line use"""
    parser = argparse.ArgumentParser(
        description="Contrast-agnostic registration using SynthSeg"
    )
    parser.add_argument(
        "--moving", required=True, help="Input moving image to be registered"
    )
    parser.add_argument(
        "--fixed", required=True, help="Reference fixed image (target space)"
    )
    parser.add_argument(
        "--skip-fixed-parc",
        action="store_true",
        help="Skip generating fixed parcellation",
    )
    parser.add_argument(
        "--skip-moving-parc",
        action="store_true",
        help="Skip generating moving parcellation",
    )
    parser.add_argument("--output", help="Output registered image")
    parser.add_argument(
        "--moving-parc", required=True, help="Path for moving image parcellation"
    )
    parser.add_argument(
        "--fixed-parc", required=True, help="Path for fixed image parcellation"
    )
    parser.add_argument(
        "--registered-parc", required=True, help="Path for registered parcellation"
    )
    parser.add_argument(
        "--affine", required=True, help="Path for affine transformation"
    )
    parser.add_argument("--warpfield", required=True, help="Path for warp field")
    parser.add_argument("--inverse-warpfield", help="Path for inverse warp field")
    parser.add_argument(
        "--generate-warpfield",
        action="store_true",
        help="Generate warp field without applying it",
    )
    parser.add_argument(
        "--apply-warpfield",
        action="store_true",
        help="Apply existing warp field to moving image",
    )
    parser.add_argument(
        "--registration-method", default="SyNRA", help="Registration method"
    )
    parser.add_argument(
        "--synthseg-threads",
        type=int,
        default=DEFAULT_THREADS,
        help="Number of threads to use for SynthSeg segmentation",
    )
    parser.add_argument(
        "--ants-threads",
        type=int,
        default=DEFAULT_THREADS,
        help="Number of threads to use for ANTs registration",
    )
    parser.add_argument("--secondary-warpfield", help="Path for secondary warp field (for robust registration)")
    parser.add_argument("--inverse-secondary-warpfield", help="Path for inverse of secondary warp field (for robust registration)")
    parser.add_argument("--qc-csv", help="Path for quality control Dice score CSV file")
    parser.add_argument("--skip-qc", action="store_true", help="Skip QC CSV generation")
    parser.add_argument("--disable-robust", action="store_true", help="Disable robust second-stage registration")
    parser.add_argument("--inverse", action="store_true", help="Whether to reverse the order of the transforms (warpfield first, then affine)")
    args = parser.parse_args()

    # Validate arguments based on workflow
    if args.apply_warpfield and (args.affine is None or args.warpfield is None):
        parser.error("--apply-warpfield requires --affine and --warpfield arguments")

    if args.generate_warpfield and args.output is not None:
        parser.error(
            "--generate-warpfield cannot be used with --output (no output image is produced)"
        )

    if not args.apply_warpfield and not args.generate_warpfield and args.output is None:
        parser.error(
            "--output is required unless --generate-warpfield or --apply-warpfield is specified"
        )

    lamareg(
        input_image=args.moving,
        reference_image=args.fixed,
        output_image=args.output,
        input_parc=args.moving_parc,
        reference_parc=args.fixed_parc,
        output_parc=args.registered_parc,
        generate_warpfield=args.generate_warpfield,
        apply_warpfield=args.apply_warpfield,
        registration_method=args.registration_method,
        affine_file=args.affine,
        warp_file=args.warpfield,
        inverse_warp_file=args.inverse_warpfield,
        synthseg_threads=args.synthseg_threads,
        ants_threads=args.ants_threads,
        qc_csv=args.qc_csv,
        skip_fixed_parc=args.skip_fixed_parc,
        skip_moving_parc=args.skip_moving_parc,
        skip_qc=args.skip_qc,
        disable_robust=args.disable_robust,
        inverse=args.inverse,
        secondary_warp_file=args.secondary_warpfield,
        inverse_secondary_warp_file=args.inverse_secondary_warpfield
    )


if __name__ == "__main__":
    main()



"""Handles loading and automatic reassembly of model files."""

import os
import shutil
from lamareg.utils.file_splitter import reassemble_file
import urllib.request

MODEL_URLS = {
    "synthseg_robust_2.0.h5": [
        "https://github.com/MICA-MNI/LAMAR-Models/raw/refs/heads/main/models/synthseg_robust_2.0.h5.000",
        "https://github.com/MICA-MNI/LAMAR-Models/raw/refs/heads/main/models/synthseg_robust_2.0.h5.001",
        "https://github.com/MICA-MNI/LAMAR-Models/raw/refs/heads/main/models/synthseg_robust_2.0.h5.002",
    ],
    "synthseg_2.0.h5": "https://github.com/MICA-MNI/LAMAR-Models/raw/refs/heads/main/models/synthseg_2.0.h5",
    "synthseg_parc_2.0.h5": "https://github.com/MICA-MNI/LAMAR-Models/raw/refs/heads/main/models/synthseg_parc_2.0.h5",
    "synthseg_qc_2.0.h5": "https://github.com/MICA-MNI/LAMAR-Models/raw/refs/heads/main/models/synthseg_qc_2.0.h5",
}


def ensure_model_file_exists(model_path):
    """Check if model file exists, and reassemble from chunks if needed.

    Args:
        model_path: Path to the model file

    Returns:
        True if the model is available (either existed or was reassembled)
    """
    if os.path.exists(model_path):
        return True

    model_name = os.path.basename(model_path)
    if model_name in MODEL_URLS:
        os.makedirs(os.path.dirname(model_path), exist_ok=True)
        print(f"Downloading {model_name}...")
        try:
            if isinstance(MODEL_URLS[model_name], list):
                # If the model is split into chunks, download each chunk
                for i, url in enumerate(MODEL_URLS[model_name]):
                    chunk_path = f"{model_path}.{i:03d}"
                    urllib.request.urlretrieve(url, chunk_path)
            else:
                urllib.request.urlretrieve(MODEL_URLS[model_name], model_path)
                return True
        except Exception as e:
            print(f"Error downloading model: {e}")
            return False

    # Model doesn't exist, look for chunks
    model_dir = os.path.dirname(model_path)
    model_name = os.path.basename(model_path)

    # Check for chunks
    chunks = [
        f
        for f in os.listdir(model_dir)
        if f.startswith(f"{model_name}.") and f[-3:].isdigit()
    ]

    if not chunks:
        print(f"ERROR: Model file {model_path} not found and no chunks detected!")
        return False

    print(
        f"Model file {model_path} not found, but {len(chunks)} chunks detected. Reassembling..."
    )
    return reassemble_file(model_path)
